#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Parametric mode class and helper functions."""
import itertools
import logging
import typing as t
from copy import deepcopy
from enum import Enum
from functools import partial

import dask.bag as db
import numpy as np
import xarray as xr
from tqdm.auto import tqdm
from typing_extensions import Literal

from pyxel.exposure import run_exposure_pipeline
from pyxel.observation.parameter_values import ParameterType, ParameterValues
from pyxel.pipelines import ResultType
from pyxel.state import get_obj_att, get_value

if t.TYPE_CHECKING:
    from pyxel.exposure import Readout
    from pyxel.outputs import ObservationOutputs
    from pyxel.pipelines import Processor


class ParameterMode(Enum):
    """Parameter mode class."""

    Product = "product"
    Sequential = "sequential"
    Custom = "custom"


class ObservationResult(t.NamedTuple):
    """Result class for observation class."""

    dataset: t.Union[xr.Dataset, t.Dict[str, xr.Dataset]]
    parameters: xr.Dataset
    logs: xr.Dataset


class Observation:
    """Observation class."""

    def __init__(
        self,
        outputs: "ObservationOutputs",
        parameters: t.Sequence[ParameterValues],
        readout: "Readout",
        mode: str = "product",
        from_file: t.Optional[str] = None,
        column_range: t.Optional[t.Tuple[int, int]] = None,
        with_dask: bool = False,
        result_type: Literal["image", "signal", "pixel", "all"] = "all",
    ):
        self.outputs = outputs
        self.readout = readout
        self.parameter_mode = ParameterMode(mode)  # type: ParameterMode
        self._parameters = parameters
        self.file = from_file
        self.data = None  # type: t.Optional[np.ndarray]
        if column_range:
            self.columns = slice(*column_range)
        self.with_dask = with_dask
        self.parameter_types = {}  # type: dict
        self._result_type = ResultType(result_type)

        if self.parameter_mode == ParameterMode.Custom:
            self._load_custom_parameters()

    def __repr__(self):
        cls_name = self.__class__.__name__  # type: str
        return f"{cls_name}<mode={self.parameter_mode!s}>"

    @property
    def result_type(self) -> ResultType:
        """TBW."""
        return self._result_type

    @result_type.setter
    def result_type(self, value: ResultType) -> None:
        """TBW."""
        self._result_type = value

    @property
    def enabled_steps(self) -> t.Sequence[ParameterValues]:
        """Return a list of enabled ParameterValues.

        Returns
        -------
        out: list
        """
        out = [step for step in self._parameters if step.enabled]
        return out

    # TODO: is self.data really needed?
    def _load_custom_parameters(self) -> None:
        """Load custom parameters from file."""
        from pyxel import load_table

        if self.file is not None:
            result = load_table(self.file).to_numpy()[
                :, self.columns
            ]  # type: np.ndarray
        else:
            raise ValueError("File for custom parametric mode not specified!")
        self.data = result

    def _custom_parameters(self) -> t.Generator[t.Tuple[int, dict], None, None]:
        """Generate custom mode parameters based on input file.

        Yields
        ------
        index: int
        parameter_dict: dict
        """
        if isinstance(self.data, np.ndarray):
            for index, data_array in enumerate(self.data):
                i = 0
                parameter_dict = {}
                for step in self.enabled_steps:
                    key = step.key

                    # TODO: this is confusing code. Fix this.
                    #       Furthermore 'step.values' should be a `t.List[int, float]` and not a `str`
                    if step.values == "_":
                        value = data_array[i]
                        i += 1
                        parameter_dict.update({key: value})

                    elif isinstance(step.values, list):

                        values = np.asarray(deepcopy(step.values))  # type: np.ndarray
                        sh = values.shape  # type: tuple
                        values_flattened = values.flatten()

                        # TODO: find a way to remove the ignore
                        if all(x == "_" for x in values_flattened):
                            value = data_array[
                                i : i + len(values_flattened)
                            ]  # noqa: E203
                            i += len(value)
                            value = value.reshape(sh).tolist()
                            parameter_dict.update({key: value})
                        else:
                            raise ValueError(
                                'Only "_" characters (or a list of them) should be used to '
                                "indicate parameters updated from file in custom mode"
                            )
                yield index, parameter_dict

        else:
            raise ValueError("Custom parameters not loaded from file.")

    def _sequential_parameters(self) -> t.Generator[t.Tuple[int, dict], None, None]:
        """Generate sequential mode parameters.

        Yields
        ------
        index: int
        parameter_dict: dict
        """
        for step in self.enabled_steps:  # type: ParameterValues
            key = step.key  # type : str
            for index, value in enumerate(step):
                parameter_dict = {key: value}
                yield index, parameter_dict

    def _product_indices(self) -> "t.Iterator[t.Tuple]":
        """Return an iterator of product parameter indices.

        Returns
        -------
        out: iterator
        """
        step_ranges = []
        for step in self.enabled_steps:
            step_ranges.append(range(len(step)))
        out = itertools.product(*step_ranges)
        return out

    def _product_parameters(
        self,
    ) -> t.Generator[t.Tuple[t.Tuple, t.Dict[str, t.Any]], None, None]:
        """Generate product mode parameters.

        Yields
        ------
        indices: tuple
        parameter_dict: dict
        """
        all_steps = self.enabled_steps
        keys = [step.key for step in self.enabled_steps]
        for indices, params in zip(
            self._product_indices(), itertools.product(*all_steps)
        ):
            parameter_dict = {}
            for key, value in zip(keys, params):
                parameter_dict.update({key: value})
            yield indices, parameter_dict

    def _parameter_it(self) -> t.Callable:
        """Return the method for generating parameters based on parametric mode.

        Returns
        -------
        callable
        """
        if self.parameter_mode == ParameterMode.Product:
            return self._product_parameters

        elif self.parameter_mode == ParameterMode.Sequential:
            return self._sequential_parameters

        elif self.parameter_mode == ParameterMode.Custom:
            return self._custom_parameters
        else:
            raise NotImplementedError

    def _processors_it(
        self, processor: "Processor"
    ) -> t.Generator[
        t.Tuple["Processor", t.Union[int, t.Tuple[int]], t.Dict], None, None
    ]:
        """Generate processors with different product parameters.

        Parameters
        ----------
        processor: Processor

        Yields
        ------
        new_processor: Processor
        index: tuple of int
        parameter_dict: dict
        """

        parameter_it = self._parameter_it()

        for index, parameter_dict in parameter_it():
            new_processor = create_new_processor(
                processor=processor, parameter_dict=parameter_dict
            )
            yield new_processor, index, parameter_dict

    def _get_parameter_types(self) -> dict:
        """Check for each step if parameters can be used as dataset coordinates (1D, simple) or not (multi).

        Returns
        -------
        self.parameter_types: dict
        """
        for step in self.enabled_steps:
            self.parameter_types.update({step.key: step.type})
        return self.parameter_types

    def run_debug_mode(
        self, processor: "Processor"
    ) -> t.Tuple[t.List["Processor"], xr.Dataset]:
        """Run obsevration pipelines in debug mode and return list of processors and parameter logs.

        Parameters
        ----------
        processor: Processor

        Returns
        -------
        processors: list
        final_logs: xr.Dataset
        """
        processors = []
        logs = []  # type: t.List[xr.Dataset]

        for processor_id, (proc, _index, parameter_dict) in enumerate(
            tqdm(self._processors_it(processor))
        ):
            log = log_parameters(
                processor_id=processor_id, parameter_dict=parameter_dict
            )  # type: xr.Dataset
            logs.append(log)
            _ = run_exposure_pipeline(
                processor=proc,
                readout=self.readout,
                outputs=self.outputs,
                progressbar=False,
            )
            processors.append(processor)

        # See issue #276
        final_logs = xr.combine_by_coords(logs)
        if not isinstance(final_logs, xr.Dataset):
            raise TypeError("Expecting 'Dataset'.")

        return processors, final_logs

    def _check_steps(self, processor: "Processor") -> None:
        """Validate enabled parameter steps in processor before running the pipelines.

        Parameters
        ----------
        processor: Processor

        Returns
        -------
        None
        """

        for step in self.enabled_steps:  # type: ParameterValues

            key = step.key  # type: str
            assert processor.has(key)

            # TODO: the string literal expressions are difficult to maintain.
            #     Example: 'pipeline.', '.arguments', '.enabled'
            #     We may want to consider an API for this.
            # Proposed API:
            # value = operator.attrgetter(step.key)(processor)
            if "pipeline." in step.key:
                model_name = step.key[: step.key.find(".arguments")]  # type: str
                model_enabled = model_name + ".enabled"  # type: str
                if not processor.get(model_enabled):
                    raise ValueError(
                        f"The '{model_name}' model referenced in parametric configuration"
                        f"has not been enabled in yaml config!"
                    )

            if (
                any(x == "_" for x in step.values[:])
                and self.parameter_mode != ParameterMode.Custom
            ):
                raise ValueError(
                    "Either define 'custom' as parametric mode or "
                    "do not use '_' character in 'values' field"
                )

    def run_observation(self, processor: "Processor") -> ObservationResult:
        """Run the observation pipelines.

        Parameters
        ----------
        processor: Processor

        Returns
        -------
        result: Result
        """
        # validation
        self._check_steps(processor)

        types = self._get_parameter_types()

        y = range(processor.detector.geometry.row)
        x = range(processor.detector.geometry.col)
        times = self.readout.times

        if self.parameter_mode == ParameterMode.Product:

            apply_pipeline = partial(
                self._apply_exposure_pipeline_product,
                x=x,
                y=y,
                processor=processor,
                times=times,
                types=types,
            )
            lst = [
                (index, parameter_dict, n)
                for n, (index, parameter_dict) in enumerate(self._parameter_it()())
            ]

            if self.with_dask:
                dataset_list = db.from_sequence(lst).map(apply_pipeline).compute()
            else:
                dataset_list = list(map(apply_pipeline, tqdm(lst)))

            # prepare lists for to-be-merged datasets
            parameters = [
                [] for _ in range(len(self.enabled_steps))
            ]  # type: t.List[t.List[xr.Dataset]]
            logs = []

            for processor_id, (indices, parameter_dict, _) in enumerate(lst):
                # log parameters for this pipeline
                log = log_parameters(
                    processor_id=processor_id, parameter_dict=parameter_dict
                )
                logs.append(log)

                # save parameters with appropriate product mode indices
                for i, coordinate in enumerate(parameter_dict):
                    parameter_ds = parameter_to_dataset(
                        parameter_dict=parameter_dict,
                        index=indices[i],
                        coordinate_name=coordinate,
                    )
                    parameters[i].append(parameter_ds)

            # merging/combining the outputs
            final_parameters_list = []  # type: t.List[xr.Dataset]
            for p in parameters:
                # See issue #276
                new_dataset = xr.combine_by_coords(p)
                if not isinstance(new_dataset, xr.Dataset):
                    raise TypeError("Expecting 'Dataset'.")

                final_parameters_list.append(new_dataset)

            final_parameters_merged = xr.merge(final_parameters_list)

            # See issue #276
            final_logs = xr.combine_by_coords(logs)
            if not isinstance(final_logs, xr.Dataset):
                raise TypeError("Expecting 'Dataset'.")

            # See issue #276
            final_dataset = xr.combine_by_coords(dataset_list)
            if not isinstance(final_dataset, xr.Dataset):
                raise TypeError("Expecting 'Dataset'.")

            result = ObservationResult(
                dataset=final_dataset,
                parameters=final_parameters_merged,
                logs=final_logs,
            )

            return result

        elif self.parameter_mode == ParameterMode.Sequential:

            apply_pipeline = partial(
                self._apply_exposure_pipeline_sequential,
                x=x,
                y=y,
                processor=processor,
                times=times,
                types=types,
            )

            lst = [
                (index, parameter_dict, n)
                for n, (index, parameter_dict) in enumerate(self._parameter_it()())
            ]

            if self.with_dask:
                dataset_list = db.from_sequence(lst).map(apply_pipeline).compute()
            else:
                dataset_list = list(map(apply_pipeline, tqdm(lst)))

            # prepare lists/dictionaries for to-be-merged datasets
            parameters = [[] for _ in range(len(self.enabled_steps))]
            logs = []

            # overflow to next parameter step counter
            step_counter = -1

            for processor_id, (index, parameter_dict, _) in enumerate(lst):
                # log parameters for this pipeline
                # TODO: somehow refactor logger so that default parameters
                #  from other steps are also logged in sequential mode
                log = log_parameters(
                    processor_id=processor_id, parameter_dict=parameter_dict
                )
                logs.append(log)

                # Figure out current coordinate
                coordinate = str(list(parameter_dict)[0])
                # Check for overflow to next parameter
                if index == 0:
                    step_counter += 1

                # save sequential parameter with appropriate index
                parameter_ds = parameter_to_dataset(
                    parameter_dict=parameter_dict,
                    index=index,
                    coordinate_name=coordinate,
                )
                parameters[step_counter].append(parameter_ds)

            # merging/combining the outputs
            # See issue #276
            final_logs = xr.combine_by_coords(logs)
            if not isinstance(final_logs, xr.Dataset):
                raise TypeError("Expecting 'Dataset'.")

            final_datasets = compute_final_sequential_dataset(
                list_of_index_and_parameter=lst, list_of_datasets=dataset_list
            )

            final_parameters_list = []
            for p in parameters:
                # See issue #276
                new_dataset = xr.combine_by_coords(p)
                if not isinstance(new_dataset, xr.Dataset):
                    raise TypeError("Expecting 'Dataset'.")

                final_parameters_list.append(new_dataset)

            final_parameters_merged = xr.merge(final_parameters_list)

            result = ObservationResult(
                dataset=final_datasets,
                parameters=final_parameters_merged,
                logs=final_logs,
            )
            return result

        elif self.parameter_mode == ParameterMode.Custom:

            apply_pipeline = partial(
                self._apply_exposure_pipeline_custom,
                x=x,
                y=y,
                processor=processor,
                times=times,
            )
            lst = [
                (index, parameter_dict, n)
                for n, (index, parameter_dict) in enumerate(self._parameter_it()())
            ]

            if self.with_dask:
                dataset_list = db.from_sequence(lst).map(apply_pipeline).compute()
            else:
                dataset_list = list(map(apply_pipeline, tqdm(lst)))

            # prepare lists for to-be-merged datasets
            logs = []

            for index, parameter_dict, _ in lst:
                # log parameters for this pipeline
                log = log_parameters(processor_id=index, parameter_dict=parameter_dict)
                logs.append(log)

            # merging/combining the outputs
            final_ds = xr.combine_by_coords(dataset_list)
            final_log = xr.combine_by_coords(logs)

            # See issue #276
            if not isinstance(final_ds, xr.Dataset):
                raise TypeError("Expecting 'Dataset'.")
            if not isinstance(final_log, xr.Dataset):
                raise TypeError("Expecting 'Dataset'.")

            final_parameters = final_log  # parameter dataset same as logs

            result = ObservationResult(
                dataset=final_ds, parameters=final_parameters, logs=final_log
            )
            return result

        else:
            raise ValueError("Parametric mode not specified.")

    def _apply_exposure_pipeline_product(
        self,
        index_and_parameter: t.Tuple[t.Tuple, t.Dict, int],
        processor: "Processor",
        x: range,
        y: range,
        times: np.ndarray,
        types: dict,
    ):

        index, parameter_dict, n = index_and_parameter

        new_processor = create_new_processor(
            processor=processor, parameter_dict=parameter_dict
        )

        # run the pipeline
        _ = run_exposure_pipeline(
            processor=new_processor,
            readout=self.readout,
            progressbar=False,
            result_type=self.result_type,
        )

        _ = self.outputs.save_to_file(processor=new_processor, run_number=n)

        ds = new_processor.result_to_dataset(
            x=x, y=y, times=times, result_type=self.result_type
        )

        # Can also be done outside dask in a loop
        ds = _add_product_parameters(
            ds=ds,
            parameter_dict=parameter_dict,
            indices=index,
            types=types,
        )

        return ds

    def _apply_exposure_pipeline_custom(
        self,
        index_and_parameter: t.Tuple[int, t.Dict, int],
        processor: "Processor",
        x: range,
        y: range,
        times: np.ndarray,
    ):

        index, parameter_dict, n = index_and_parameter

        new_processor = create_new_processor(
            processor=processor, parameter_dict=parameter_dict
        )

        # run the pipeline
        _ = run_exposure_pipeline(
            processor=new_processor,
            readout=self.readout,
            progressbar=False,
            result_type=self.result_type,
        )

        _ = self.outputs.save_to_file(processor=new_processor, run_number=n)

        ds = new_processor.result_to_dataset(
            x=x, y=y, times=times, result_type=self.result_type
        )

        # Can also be done outside dask in a loop
        ds = _add_custom_parameters(
            ds=ds,
            index=index,
        )

        return ds

    def _apply_exposure_pipeline_sequential(
        self,
        index_and_parameter: t.Tuple[int, t.Dict, int],
        processor: "Processor",
        x: range,
        y: range,
        times: np.ndarray,
        types: dict,
    ):

        index, parameter_dict, n = index_and_parameter

        new_processor = create_new_processor(
            processor=processor, parameter_dict=parameter_dict
        )

        coordinate = str(list(parameter_dict)[0])

        # run the pipeline
        _ = run_exposure_pipeline(
            processor=new_processor,
            readout=self.readout,
            progressbar=False,
            result_type=self.result_type,
        )

        _ = self.outputs.save_to_file(processor=new_processor, run_number=n)

        ds = new_processor.result_to_dataset(
            x=x, y=y, times=times, result_type=self.result_type
        )

        # Can also be done outside dask in a loop
        ds = _add_sequential_parameters(
            ds=ds,
            parameter_dict=parameter_dict,
            index=index,
            coordinate_name=coordinate,
            types=types,
        )

        return ds

    def debug_parameters(self, processor: "Processor") -> list:
        """List the parameters using processor parameters in processor generator.

        Parameters
        ----------
        processor: Processor

        Returns
        -------
        result: list
        """
        result = []
        processor_generator = self._processors_it(processor=processor)
        for i, (proc, _, _) in enumerate(processor_generator):
            values = []
            for step in self.enabled_steps:
                _, att = get_obj_att(proc, step.key)
                value = get_value(proc, step.key)
                values.append((att, value))
            logging.debug("%d: %r" % (i, values))
            result.append((i, values))
        return result


def create_new_processor(processor: "Processor", parameter_dict: dict) -> "Processor":
    """Create a copy of processor and set new attributes from a dictionary before returning it.

    Parameters
    ----------
    processor: Processor
    parameter_dict: dict

    Returns
    -------
    new_processor: Processor
    """

    new_processor = deepcopy(processor)

    for key in parameter_dict.keys():
        new_processor.set(key=key, value=parameter_dict[key])

    return new_processor


def _id(s: str) -> str:
    """Add _id to the end of a string.

    Parameters
    ----------
    s: str

    Returns
    -------
    out: str
    """
    out = s + "_id"
    return out


def short(s: str) -> str:
    """Split string with . and return the last element.

    Parameters
    ----------
    s: str

    Returns
    -------
    out: str
    """
    out = s.split(".")[-1]
    return out


def log_parameters(processor_id: int, parameter_dict: dict) -> xr.Dataset:
    """Return parameters in the current processor in a xarray dataset.

    Parameters
    ----------
    processor_id: int
    parameter_dict: dict

    Returns
    -------
    out: xr.Dataset
    """
    out = xr.Dataset()
    for key, value in parameter_dict.items():
        da = xr.DataArray(value)
        da = da.assign_coords(coords={"id": processor_id})
        da = da.expand_dims(dim="id")
        out[short(key)] = da
    return out


def parameter_to_dataset(
    parameter_dict: dict, index: int, coordinate_name: str
) -> xr.Dataset:
    """Return a specific parameter dataset from a parameter dictionary.

    Parameters
    ----------
    parameter_dict: dict
    index: int
    coordinate_name: str

    Returns
    -------
    parameter_ds: xr.Dataset
    """

    parameter_ds = xr.Dataset()
    parameter = xr.DataArray(parameter_dict[coordinate_name])
    parameter = parameter.assign_coords({short(_id(coordinate_name)): index})
    parameter = parameter.expand_dims(dim=short(_id(coordinate_name)))
    parameter_ds[short(coordinate_name)] = parameter

    return parameter_ds


def _add_custom_parameters(ds: xr.Dataset, index: int) -> xr.Dataset:
    """Add coordinate coordinate "index" to the dataset.

    Parameters
    ----------
    ds: xarray.Dataset
    index: ind

    Returns
    -------
    ds: xr.Dataset
    """

    ds = ds.assign_coords({"id": index})
    ds = ds.expand_dims(dim="id")

    return ds


def _add_sequential_parameters(
    ds: xr.Dataset,
    parameter_dict: dict,
    index: int,
    coordinate_name: str,
    types: dict,
) -> xr.Dataset:
    """Add true coordinates or index to sequential mode dataset.

    Parameters
    ----------
    ds: xr.Dataset
    parameter_dict: dict
    index: int
    coordinate_name: str
    types: dict

    Returns
    -------
    ds: xr.Dataset
    """

    #  assigning the right coordinates based on type
    if types[coordinate_name] == ParameterType.Simple:
        ds = ds.assign_coords(
            coords={short(coordinate_name): parameter_dict[coordinate_name]}
        )
        ds = ds.expand_dims(dim=short(coordinate_name))

    elif types[coordinate_name] == ParameterType.Multi:
        ds = ds.assign_coords({short(_id(coordinate_name)): index})
        ds = ds.expand_dims(dim=short(_id(coordinate_name)))

    return ds


def _add_product_parameters(
    ds: xr.Dataset, parameter_dict: dict, indices: t.Tuple, types: dict
) -> xr.Dataset:
    """Add true coordinates or index to product mode dataset.

    Parameters
    ----------
    ds: xr.Dataset
    parameter_dict: dict
    indices: tuple
    types: dict

    Returns
    -------
    ds: xr.Dataset
    """

    for i, coordinate in enumerate(parameter_dict):

        #  assigning the right coordinates based on type
        if types[coordinate] == ParameterType.Simple:
            ds = ds.assign_coords(
                coords={short(coordinate): parameter_dict[coordinate]}
            )
            ds = ds.expand_dims(dim=short(coordinate))

        elif types[coordinate] == ParameterType.Multi:
            ds = ds.assign_coords({short(_id(coordinate)): indices[i]})
            ds = ds.expand_dims(dim=short(_id(coordinate)))

        else:
            raise NotImplementedError

    return ds


def compute_final_sequential_dataset(
    list_of_index_and_parameter: list, list_of_datasets: list
) -> t.Dict[str, xr.Dataset]:
    """Return a dictionary of result datasets where keys are different parameters.

    Parameters
    ----------
    list_of_index_and_parameter: list
    list_of_datasets: list

    Returns
    -------
    final_datasets: dict
    """

    final_dict = {}  # type: t.Dict[str, list]

    for _, parameter_dict, n in list_of_index_and_parameter:
        coordinate = str(list(parameter_dict)[0])
        if short(coordinate) not in final_dict.keys():
            final_dict.update({short(coordinate): []})
            final_dict[short(coordinate)].append(list_of_datasets[n])
        else:
            final_dict[short(coordinate)].append(list_of_datasets[n])

    final_datasets = {}  # type: t.Dict[str, xr.Dataset]
    for key, value in final_dict.items():
        ds = xr.combine_by_coords(value)
        # see issue #276
        if not isinstance(ds, xr.Dataset):
            raise TypeError("Expecting 'Dataset'.")

        final_datasets.update({key: ds})

    return final_datasets



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

import typing as t

import numpy as np

from pyxel.detectors import CCD

try:
    import arcticpy as ac

    WITH_ARTICPY = True
except ImportError:
    # No 'arcticpy' library
    WITH_ARTICPY = False


def arctic_add(
    detector: CCD,
    well_fill_power: float,
    trap_densities: list,
    trap_release_timescales: list,
    express: int = 0,
    # instant_traps: t.Sequence[t.Mapping[str, float]],
) -> None:
    """Add trap species.

    Parameters
    ----------
    detector : CCD
    well_fill_power
    trap_densities
    trap_release_timescales
    express
    """
    if not WITH_ARTICPY:
        raise RuntimeError(
            "ArCTIC python wrapper is not installed ! "
            "See https://github.com/jkeger/arctic"
        )

    ccd = ac.CCD(
        phases=[
            ac.CCDPhase(
                full_well_depth=detector.characteristics.fwc,
                well_fill_power=well_fill_power,
            )
        ]
    )

    roe = ac.ROE()

    traps = []  # type: t.List[ac.TrapInstantCapture]
    i = 0
    for trap_density in trap_densities:

        instant_traps = [
            {"density": trap_density, "release_timescale": trap_release_timescales[i]}
        ]  # type: t.Sequence[t.Mapping[str, float]]

        # Build the traps
        for trap_info in instant_traps:
            density = trap_info["density"]  # type: float
            release_timescale = trap_info["release_timescale"]  # type: float
            trap = ac.TrapInstantCapture(
                density=density, release_timescale=release_timescale
            )
            traps.append(trap)
        # Add CTI
        i += 1

    image_2d = np.asarray(detector.pixel.array, dtype=float)  # type: np.ndarray

    image_cti_added_2d = ac.add_cti(
        image=image_2d,
        parallel_traps=traps,
        parallel_ccd=ccd,
        parallel_roe=roe,
        parallel_express=express,
        verbosity=0,
    )

    detector.pixel.array = image_cti_added_2d


def arctic_remove(
    detector: CCD,
    well_fill_power: float,
    instant_traps: t.Sequence[t.Mapping[str, float]],
    num_iterations: int,
    express: int = 0,
) -> None:
    """Remove trap species.

    Parameters
    ----------
    detector
    well_fill_power
    instant_traps
    num_iterations
    express
    """
    if not WITH_ARTICPY:
        raise RuntimeError(
            "ArCTIC python wrapper is not installed ! "
            "See https://github.com/jkeger/arctic"
        )

    ccd = ac.CCD(
        well_fill_power=well_fill_power, full_well_depth=detector.characteristics.fwc
    )
    roe = ac.ROE()

    # Build the traps
    traps = []  # type: t.List[ac.Trap]
    for trap_info in instant_traps:
        density = trap_info["density"]  # type: float
        release_timescale = trap_info["release_timescale"]  # type: float

        trap = ac.Trap(density=density, release_timescale=release_timescale)
        traps.append(trap)

    # Remove CTI
    image_2d = np.asarray(detector.pixel.array, dtype=float)  # type: np.ndarray

    image_cti_removed = ac.remove_cti(
        image=image_2d,
        iterations=num_iterations,
        parallel_traps=traps,
        parallel_ccd=ccd,
        parallel_roe=roe,
        parallel_express=express,
    )

    detector.pixel.array = image_cti_removed



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel photon generator models."""
from pyxel.data_structure import Photon
from pyxel.detectors import Detector


# @pyxel.validate
# @pyxel.argument(name='image_file', label='', validate=check_path)
def alignment(detector: Detector) -> None:
    """Optical alignment.

    :param detector: Pyxel Detector object
    """
    geo = detector.geometry
    rows, cols = detector.photon.array.shape
    row0 = int((rows - geo.row) / 2)
    col0 = int((cols - geo.col) / 2)

    if row0 < 0 or col0 < 0:
        raise ValueError

    aligned_optical_image = detector.photon.array[
        slice(row0, row0 + geo.row), slice(col0, col0 + geo.col)
    ]

    detector.photon = Photon(aligned_optical_image)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from collections import abc

import pytest

from pyxel.models.photon_generation import illumination
from pyxel.pipelines import Arguments, ModelFunction


@pytest.fixture
def model_function() -> ModelFunction:
    """Create a valid `ModelFunction` instance."""
    return ModelFunction(
        name="illumination",
        func=illumination,
        arguments={"level": 1, "option": "foo"},
    )


def test_type_model_function(model_function: ModelFunction):
    assert isinstance(model_function, ModelFunction)
    assert callable(model_function)


@pytest.mark.skip(reason="Class `Arguments` is not implemented")
def test_type_arguments(model_function: ModelFunction):
    assert isinstance(model_function.arguments, Arguments)
    assert isinstance(model_function.arguments, abc.Mapping)


@pytest.mark.parametrize("key", ["level", "option"])
def test_arguments_contain(model_function, key):
    assert isinstance(model_function, ModelFunction)
    assert key in model_function.arguments


@pytest.mark.parametrize("key", ["", "LEVEL", "foo"])
def test_arguments_not_contains(model_function, key):
    assert isinstance(model_function, ModelFunction)
    assert key not in model_function.arguments


@pytest.mark.parametrize("key, exp_value", [("level", 1), ("option", "foo")])
def test_arguments_getitem(model_function, key, exp_value):
    assert model_function.arguments[key] == exp_value


def test_arguments_iter(model_function):
    assert list(model_function.arguments) == ["level", "option"]


def test_arguments_len(model_function):
    assert len(model_function.arguments) == 2


@pytest.mark.skip(reason="Class `Arguments` is not implemented")
def test_arguments_getattr(model_function):
    assert model_function.arguments.level == 1


def test_arguments_getattr_error(model_function):
    with pytest.raises(AttributeError):
        _ = model_function.arguments.Level


@pytest.mark.skip(reason="Class `Arguments` is not implemented")
def test_get_level(model_function):
    assert isinstance(model_function, ModelFunction)

    result = model_function.arguments.level
    assert result == 1



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import pytest

try:
    import pygmo as pg

    WITH_PYGMO = True
except ImportError:
    WITH_PYGMO = False

import pyxel
from pyxel import Configuration
from pyxel.calibration import Algorithm, Calibration, CalibrationMode
from pyxel.data_structure import Charge, Image, Pixel, Signal
from pyxel.detectors import CCD, CCDCharacteristics, CCDGeometry, Environment, Material
from pyxel.exposure import Exposure
from pyxel.observation import Observation, ParameterMode
from pyxel.outputs.calibration_outputs import CalibrationOutputs
from pyxel.outputs.exposure_outputs import ExposureOutputs
from pyxel.outputs.observation_outputs import ObservationOutputs
from pyxel.pipelines import DetectionPipeline, ModelFunction, ModelGroup


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml_file",
    [
        "tests/data/parametric.yaml",
        "tests/data/yaml.yaml",
        "tests/data/calibrate_models.yaml",
    ],
)
def test_yaml_load(yaml_file):
    cfg = pyxel.load(yaml_file)

    assert isinstance(cfg, Configuration)

    if isinstance(cfg.exposure, Exposure):
        assert isinstance(cfg.exposure.outputs, ExposureOutputs)
    elif isinstance(cfg.calibration, Calibration):
        assert isinstance(cfg.calibration.outputs, CalibrationOutputs)
        assert isinstance(cfg.calibration.algorithm, Algorithm)
        assert isinstance(cfg.calibration.calibration_mode, CalibrationMode)
    elif isinstance(cfg.observation, Observation):
        assert isinstance(cfg.observation.outputs, ObservationOutputs)
        assert isinstance(cfg.observation.parameter_mode, ParameterMode)
    else:
        raise AssertionError("Running mode not initialized.")

    assert isinstance(cfg.ccd_detector, CCD)
    assert isinstance(cfg.ccd_detector.geometry, CCDGeometry)
    assert isinstance(cfg.ccd_detector.characteristics, CCDCharacteristics)
    assert isinstance(cfg.ccd_detector.material, Material)
    assert isinstance(cfg.ccd_detector.environment, Environment)
    assert isinstance(cfg.ccd_detector.image, Image)
    assert isinstance(cfg.ccd_detector.signal, Signal)
    assert isinstance(cfg.ccd_detector.pixel, Pixel)
    assert isinstance(cfg.ccd_detector.charge, Charge)

    assert isinstance(cfg.pipeline, DetectionPipeline)
    assert isinstance(cfg.pipeline.photon_generation, ModelGroup)
    assert isinstance(cfg.pipeline.photon_generation.models[0], ModelFunction)
    assert isinstance(cfg.pipeline.charge_generation, ModelGroup)
    assert isinstance(cfg.pipeline.charge_generation.models[0], ModelFunction)
    assert isinstance(cfg.pipeline.charge_collection, ModelGroup)
    assert isinstance(cfg.pipeline.charge_collection.models[0], ModelFunction)
    assert isinstance(cfg.pipeline.charge_transfer, ModelGroup)
    assert isinstance(cfg.pipeline.charge_transfer.models[0], ModelFunction)
    assert isinstance(cfg.pipeline.charge_measurement, ModelGroup)
    assert isinstance(cfg.pipeline.charge_measurement.models[0], ModelFunction)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel detector simulation framework."""

# flake8: noqa
from ._version import get_versions

__version__ = get_versions()["version"]  # type: str
del get_versions

from .options import SetOptions as set_options
from .show_versions import show_versions
from .inputs import load_image, load_table, load_datacube
from .configuration import load, save, Configuration
from .run import calibration_mode, exposure_mode, observation_mode, run
from .notebook import (
    display_detector,
    display_persist,
    display_html,
    display_calibration_inputs,
    display_simulated,
    display_evolution,
    champion_heatmap,
    optimal_parameters,
)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

# flake8: noqa
from .algorithm import AlgorithmType, Algorithm
from .user_defined import DaskBFE, DaskIsland
from .protocols import IslandProtocol, ProblemSingleObjective
from .util import (
    CalibrationResult,
    CalibrationMode,
    Island,
    check_ranges,
    list_to_slice,
    read_data,
    read_datacubes,
    list_to_3d_slice,
)
from ..pipelines.processor import ResultType
from .archipelago import MyArchipelago
from .calibration import Calibration, CalibrationMode
from .fitness import sum_of_abs_residuals, sum_of_squared_residuals
from .fitting import ModelFitting



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Geometry class for detector."""
import logging
import typing as t
from enum import Enum
from pathlib import Path

import numpy as np

from pyxel.util.memory import get_size


class MaterialType(Enum):
    """TBW."""

    Silicon = "silicon"
    HXRG = "hxrg"


# Universal global constants
M_ELECTRON = 9.10938356e-31  # kg     # TODO put these global constants to a data file


def load_array(filename: Path) -> np.ndarray:
    """Load an array from a filename."""
    if filename.suffix == ".npy":
        logging.debug("Load numpy array from filename '%s'", filename)
        result = np.load(filename)  # type: np.ndarray
    else:
        raise NotImplementedError

    return result


class Material:
    """Material attributes of the detector."""

    def __init__(
        self,
        trapped_charge: t.Optional[Path] = None,
        n_acceptor: float = 0.0,
        n_donor: float = 0.0,
        material: MaterialType = MaterialType.Silicon,
        material_density: float = 2.328,
        ionization_energy: float = 3.6,
        band_gap: float = 1.12,
        e_effective_mass: float = 0.5 * M_ELECTRON,
    ):
        """Create an instance of `Material`.

        Parameters
        ----------
        trapped_charge: str
            Numpy array storing the trap density temporarily.
        n_acceptor: float
            Density of acceptors in the lattice. Unit: cm^-3
        n_donor: float
            Density of donors in the lattice. Unit: cm^-3
        material: MaterialType
            Semiconductor material of the detector
        material_density: float
            Material density. Unit: g/cm^3
        ionization_energy: float
            Mean ionization energy of the semiconductor lattice. Unit: eV
        band_gap: float
            Band gap of the semiconductor lattice. Unit: eV
        e_effective_mass: float
            Electron effective mass in the semiconductor lattice. Unit: kg
        """
        if not (0.0 <= n_acceptor <= 1000.0):
            raise ValueError("'n_acceptor' must be between 0.0 and 1000.0.")

        if not (0.0 <= n_donor <= 1000.0):
            raise ValueError("'n_donor' must be between 0.0 and 1000.0.")

        if not (0.0 <= material_density <= 10000.0):
            raise ValueError("'material_density' must be between 0.0 and 10000.0.")

        if not (0.0 <= ionization_energy <= 100.0):
            raise ValueError("'ionization_energy' must be between 0.0 and 100.0.")

        if not (0.0 <= band_gap <= 10.0):
            raise ValueError("'band_gap' must be between 0.0 and 10.0.")

        if not (0.0 <= e_effective_mass <= 1.0e-10):
            raise ValueError("'e_effective_mass' must be between 0.0 and 1.e-10.")

        data = (
            load_array(trapped_charge) if trapped_charge else None
        )  # type: t.Optional[np.ndarray]

        self._trapped_charge = data  # type: t.Optional[np.ndarray]
        self._n_acceptor = n_acceptor
        self._n_donor = n_donor

        # TODO: The following parameters could be extracted into a new dedicated class
        self._material = material  # TODO: create func for compound materials
        self._material_density = (
            material_density  # TODO: set automatically depending on the material
        )
        self._ionization_energy = (
            ionization_energy  # TODO: set automatically depending on the material
        )
        self._band_gap = band_gap  # TODO: set automatically depending on the material
        self._e_effective_mass = (
            e_effective_mass  # TODO: set automatically depending on the material
        )
        self._numbytes = 0

    @property
    def trapped_charge(self) -> t.Optional[np.ndarray]:
        """Get Numpy array storing the trap density temporarily."""
        return self._trapped_charge

    @trapped_charge.setter
    def trapped_charge(self, value: Path) -> None:
        """Set Numpy array storing the trap density temporarily."""
        self._trapped_charge = load_array(value)

    @property
    def n_acceptor(self) -> float:
        """Get Density of acceptors in the lattice."""
        return self._n_acceptor

    @n_acceptor.setter
    def n_acceptor(self, value: float) -> None:
        """Set Density of acceptors in the lattice."""
        if not (0.0 <= value <= 1000.0):
            raise ValueError("'n_acceptor' must be between 0.0 and 1000.0.")

        self._n_acceptor = value

    @property
    def n_donor(self) -> float:
        """Get Density of donors in the lattice."""
        return self._n_donor

    @n_donor.setter
    def n_donor(self, value: float) -> None:
        """Set Density of donors in the lattice."""
        if not (0.0 <= value <= 1000.0):
            raise ValueError("'n_donor' must be between 0.0 and 1000.0.")

        self._n_donor = value

    @property
    def material(self) -> MaterialType:
        """Get Semiconductor material of the detector."""
        return self._material

    @material.setter
    def material(self, value: MaterialType) -> None:
        """Set Semiconductor material of the detector."""
        self._material = value

    @property
    def material_density(self) -> float:
        """Get Material density."""
        return self._n_donor

    @material_density.setter
    def material_density(self, value: float) -> None:
        """Set Material density."""
        if not (0.0 <= value <= 10000.0):
            raise ValueError("'material_density' must be between 0.0 and 10000.0.")

        self._material_density = value

    @property
    def ionization_energy(self) -> float:
        """Get Mean ionization energy of the semiconductor lattice."""
        return self._ionization_energy

    @ionization_energy.setter
    def ionization_energy(self, value: float) -> None:
        """Set Mean ionization energy of the semiconductor lattice."""
        if not (0.0 <= value <= 100.0):
            raise ValueError("'ionization_energy' must be between 0.0 and 100.0.")

        self._ionization_energy = value

    @property
    def band_gap(self) -> float:
        """Get Band gap of the semiconductor lattice."""
        return self._band_gap

    @band_gap.setter
    def band_gap(self, value: float) -> None:
        """Set Band gap of the semiconductor lattice."""
        if not (0.0 <= value <= 10.0):
            raise ValueError("'band_gap' must be between 0.0 and 10.0.")

        self._band_gap = value

    @property
    def e_effective_mass(self) -> float:
        """Get Electron effective mass in the semiconductor lattice."""
        return self._e_effective_mass

    @e_effective_mass.setter
    def e_effective_mass(self, value: float) -> None:
        """Set Electron effective mass in the semiconductor lattice."""
        if not (0.0 <= value <= 1.0e-10):
            raise ValueError("'e_effective_mass' must be between 0.0 and 1.e-10.")

        self._e_effective_mass = value

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes

    # TODO create func for compound materials
    # def set_material(self, material):
    #     """Set material properties.
    #
    #     :param material:
    #     """
    #     # TODO put these constants to a data file
    #     if material == 'silicon' or 'Si' or 'si':
    #         self.material_density = 2.328                 # (g/cm3)
    #         self.ionization_energy = 3.6                  # (eV)
    #         self.band_gap = 1.12                          # (eV)
    #         self.e_effective_mass = 0.5 * M_ELECTRON      # (kg)
    #
    #     else:
    #         raise NotImplementedError('Given material has not implemented yet')



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Sub-package to create 'archipelagos'."""
import logging
import typing as t
from concurrent.futures.thread import ThreadPoolExecutor
from random import Random
from timeit import default_timer as timer

import dask.array as da
import dask.delayed as delayed
import numpy as np
import pandas as pd
import xarray as xr
from tqdm.auto import tqdm

from pyxel.calibration import Algorithm, AlgorithmType, IslandProtocol
from pyxel.calibration.fitting import ModelFitting

try:
    import pygmo as pg
except ImportError:
    pass

if t.TYPE_CHECKING:
    from numpy.typing import ArrayLike

    from pyxel.exposure import Readout


class ArchipelagoLogs:
    """Keep log information from all algorithms in an archipelago."""

    def __init__(self, algo_type: AlgorithmType, num_generations: int):
        self._df = pd.DataFrame()
        self._num_generations = num_generations

        # Get logging's columns that will be extracted from the algorithm
        if algo_type is AlgorithmType.Sade:
            self._columns = (
                "num_generations",  # Generation number
                "num_evaluations",  # Number of functions evaluation made
                "best_fitness",  # The best fitness currently in the population
                "f",
                "cr",
                "dx",
                "df",
            )  # See: https://esa.github.io/pygmo2/algorithms.html#pygmo.sade.get_log
            self._algo_to_extract = pg.sade
        else:
            raise NotImplementedError

    def _from_algo(self, algo: "pg.algorithm") -> pd.DataFrame:
        """Get logging information from an algorithm."""
        # Extract the Pygmo algorithm and its logging information
        algo_extracted = algo.extract(self._algo_to_extract)
        logs = algo_extracted.get_log()  # type: list

        # Put the logging information from pygmo into a `DataFrame`
        df = pd.DataFrame(logs, columns=self._columns)

        return df

    def _from_archi(self, archi: "pg.archipelago") -> pd.DataFrame:
        """Get logging information from an archipelago."""
        lst = []  # type: t.List[pd.DataFrame]

        for id_island, island in enumerate(archi):
            # Extract the Pygmo algorithm from an island
            df_island = self._from_algo(algo=island.get_algorithm())
            df_island["id_island"] = id_island + 1

            lst.append(df_island)

        df_archipelago = pd.concat(lst)
        return df_archipelago

    # TODO: Remove parameter 'id_evolution' ?
    def append(self, archi: "pg.archipelago", id_evolution: int) -> None:
        """Collect logging information from a archipelago for specified evolution id."""
        partial_df = self._from_archi(archi=archi)  # type: pd.DataFrame
        partial_df["id_evolution"] = id_evolution

        self._df = self._df.append(partial_df)

    def get_full_total(self) -> pd.DataFrame:
        """TBW."""
        new_df = self._df.reset_index(drop=True)
        new_df["global_num_generations"] = (
            new_df["id_evolution"] - 1
        ) * self._num_generations + new_df["num_generations"]

        return new_df

    def get_total(self) -> pd.DataFrame:
        """TBW."""
        df = self.get_full_total()  # type: pd.DataFrame
        return df[
            ["id_evolution", "id_island", "global_num_generations", "best_fitness"]
        ]


# def extract_data_2d(df_processors: pd.DataFrame, rows: int, cols: int) -> xr.Dataset:
#     """Extract 'image', 'signal' and 'pixel' arrays from several delayed processors.
#
#     Parameters
#     ----------
#     df_processors : DataFrame
#         A dataframe with the columns 'island', 'id_processor' and 'processor'.
#         Data under column 'processor' are ``Delayed`` ``Processor`` objects.
#     rows : int
#         rows and cols are extracted from the geometry of the processor
#     cols : int
#
#     Returns
#     -------
#     Dataset
#         TBW.
#
#     Examples
#     --------
#     >>> df
#         island  id_processor                                          processor
#     0        0             0  Delayed('apply_parameters-f74290ab-874e-44b3-b...
#     2        0             1  Delayed('apply_parameters-467369d7-ea31-497e-a...
#     4        0             2  Delayed('apply_parameters-5bebe432-128d-406c-b...
#     6        0             3  Delayed('apply_parameters-28de0355-77c4-4a4d-b...
#     8        0             4  Delayed('apply_parameters-dbe2b999-8d03-416b-b...
#     10       0             5  Delayed('apply_parameters-482ec044-6245-42a5-9...
#     1        1             0  Delayed('apply_parameters-8e0e9828-740a-4292-9...
#     3        1             1  Delayed('apply_parameters-e1fe8b45-41da-4304-a...
#     5        1             2  Delayed('apply_parameters-910ff196-e8c9-41ad-8...
#     7        1             3  Delayed('apply_parameters-fa6c5430-752a-49b1-8...
#     9        1             4  Delayed('apply_parameters-9a29b340-f0e0-48d0-8...
#     11       1             5  Delayed('apply_parameters-5b8cf0e1-d62f-4fcb-9...
#
#     >>> rows, cols
#     (835, 1)
#     >>> extract_data_2d(df_processors=df, rows=rows, cols=cols)
#     <xarray.Dataset>
#     Dimensions:           (id_processor: 6, island: 2, x: 1, y: 835)
#     Coordinates:
#       * island            (island) int64 0 1
#       * id_processor      (id_processor) int64 0 1 2 3 4 5
#       * y                 (y) int64 0 1 2 3 4 5 6 7 ... 828 829 830 831 832 833 834
#       * x                 (x) int64 0
#     Data variables:
#         simulated_image   (island, id_processor, y, x) float64 dask.array<chunksize=(1, 1, 835, 1), meta=np.ndarray>
#         simulated_signal  (island, id_processor, y, x) float64 dask.array<chunksize=(1, 1, 835, 1), meta=np.ndarray>
#         simulated_pixel   (island, id_processor, y, x) float64 dask.array<chunksize=(1, 1, 835, 1), meta=np.ndarray>
#     """
#     lst = []
#     for _, row in df_processors.iterrows():
#         island = row["island"]  # type: int
#         id_processor = row["id_processor"]  # type: int
#         processor = row["processor"]  # type: delayed.Delayed
#
#         image_delayed = processor.detector.image.array  # type: delayed.Delayed
#         signal_delayed = processor.detector.signal.array  # type: delayed.Delayed
#         pixel_delayed = processor.detector.pixel.array  # type: delayed.Delayed
#
#         image_2d = da.from_delayed(image_delayed, shape=(rows, cols), dtype=float)
#         signal_2d = da.from_delayed(signal_delayed, shape=(rows, cols), dtype=float)
#         pixel_2d = da.from_delayed(pixel_delayed, shape=(rows, cols), dtype=float)
#
#         partial_ds = xr.Dataset()
#         partial_ds["simulated_image"] = xr.DataArray(image_2d, dims=["y", "x"])
#         partial_ds["simulated_signal"] = xr.DataArray(signal_2d, dims=["y", "x"])
#         partial_ds["simulated_pixel"] = xr.DataArray(pixel_2d, dims=["y", "x"])
#
#         lst.append(
#             partial_ds.assign_coords(
#                 island=island,
#                 id_processor=id_processor,
#             ).expand_dims(["island", "id_processor"])
#         )
#
#     ds = xr.combine_by_coords(lst).assign_coords(
#         y=range(rows),
#         x=range(cols),
#     )  # type: xr.Dataset
#
#     return ds


def extract_data_3d(
    df_results: pd.DataFrame,
    rows: int,
    cols: int,
    times: int,
    readout_times: np.ndarray,
) -> xr.Dataset:
    """Extract 'image', 'signal' and 'pixel' arrays from several delayed dynamic results."""
    lst = []
    for _, row in df_results.iterrows():
        island = row["island"]  # type: int
        id_processor = row["id_processor"]  # type: int
        result = row["processor"].result  # type: delayed.Delayed

        image_delayed = result["image"]  # type: delayed.Delayed
        signal_delayed = result["signal"]  # type: delayed.Delayed
        pixel_delayed = result["pixel"]  # type: delayed.Delayed

        image_3d = da.from_delayed(
            image_delayed, shape=(times, rows, cols), dtype=float
        )
        signal_3d = da.from_delayed(
            signal_delayed, shape=(times, rows, cols), dtype=float
        )
        pixel_3d = da.from_delayed(
            pixel_delayed, shape=(times, rows, cols), dtype=float
        )

        partial_ds = xr.Dataset()
        partial_ds["simulated_image"] = xr.DataArray(
            image_3d, dims=["readout_time", "y", "x"]
        )
        partial_ds["simulated_signal"] = xr.DataArray(
            signal_3d, dims=["readout_time", "y", "x"]
        )
        partial_ds["simulated_pixel"] = xr.DataArray(
            pixel_3d, dims=["readout_time", "y", "x"]
        )

        lst.append(
            partial_ds.assign_coords(
                island=island,
                id_processor=id_processor,
            ).expand_dims(["island", "id_processor"])
        )

    ds = xr.combine_by_coords(lst).assign_coords(
        readout_time=readout_times,
        y=range(rows),
        x=range(cols),
    )  # type: xr.Dataset

    return ds


# TODO: Rename to PyxelArchipelago
class MyArchipelago:
    """User-defined Archipelago."""

    def __init__(
        self,
        num_islands: int,
        udi: IslandProtocol,
        algorithm: Algorithm,
        problem: ModelFitting,
        pop_size: int,
        bfe: t.Optional[t.Callable] = None,
        topology: t.Optional[t.Callable] = None,
        seed: t.Optional[int] = None,
        parallel: bool = True,
        with_bar: bool = False,
    ):
        self._log = logging.getLogger(__name__)

        self.num_islands = num_islands
        self.udi = udi
        self.algorithm = algorithm  # type: Algorithm
        self.problem = problem  # type: ModelFitting
        self.pop_size = pop_size
        self.bfe = bfe
        self.topology = topology
        self.seed = seed
        self.parallel = parallel
        self.with_bar = with_bar

        # Create a Pygmo archipelago
        self._pygmo_archi = pg.archipelago(t=self.topology)

        # Create a Pygmo algorithm
        verbosity_level = max(1, self.algorithm.population_size // 100)  # type: int

        self._pygmo_algo = pg.algorithm(
            self.algorithm.get_algorithm()
        )  # type: pg.algorithm
        self._pygmo_algo.set_verbosity(verbosity_level)
        self._log.info(self._pygmo_algo)

        # Create a Pygmo problem
        self._pygmo_prob = pg.problem(self.problem)  # type: pg.problem
        self._log.info(self._pygmo_prob)

        # Build the archipelago
        self._build()

    def _build(self) -> None:
        """Build the island(s) and populate them."""
        disable_bar = not self.with_bar  # type: bool
        start_time = timer()  # type: float

        def create_island(seed: t.Optional[int] = None) -> pg.island:
            """Create a new island."""
            return pg.island(
                udi=self.udi,
                algo=self._pygmo_algo,
                prob=self._pygmo_prob,
                b=self.bfe,
                size=self.pop_size,
                seed=seed,
            )

        # Create a seed for each island
        if self.seed is None:
            seeds = [None] * self.num_islands  # type: t.Sequence[t.Optional[int]]
        else:
            func_rnd = Random()  # type: Random
            func_rnd.seed(self.seed)
            max_value = np.iinfo(np.uint32).max  # type: int
            seeds = [func_rnd.randint(0, max_value) for _ in range(self.num_islands)]

        # Create the islands and add them to this archipelago
        if self.parallel:
            # Create the islands in parallel with Threads
            with ThreadPoolExecutor(max_workers=self.num_islands) as executor:
                it = executor.map(create_island, seeds)

                for island in tqdm(
                    it,
                    desc="Create islands",
                    total=self.num_islands,
                    disable=disable_bar,
                ):
                    self._pygmo_archi.push_back(island)
        else:
            # Create the islands sequentially
            it = map(create_island, seeds)
            for island in tqdm(
                it, desc="Create islands", total=self.num_islands, disable=disable_bar
            ):
                self._pygmo_archi.push_back(island)

        stop_time = timer()
        logging.info("Create a new archipelago in %.2f s", stop_time - start_time)

    def run_evolve(
        self,
        readout: "Readout",
        num_evolutions: int = 1,
        num_best_decisions: t.Optional[int] = None,
    ) -> t.Tuple[xr.Dataset, pd.DataFrame, pd.DataFrame]:
        """Run evolution(s) several time.

        Parameters
        ----------
        sampling
        num_evolutions : int
            Number of time to run the evolutions.
        num_best_decisions : int or None, optional.
            Number of best individuals to extract. If this parameter is set to None then
            no individuals are extracted.

        Returns
        -------
        TBW.
        """
        self._log.info("Run %i evolutions", num_evolutions)
        logs = ArchipelagoLogs(
            algo_type=self.algorithm.type, num_generations=self.algorithm.generations
        )

        total_num_generations = num_evolutions * self.algorithm.generations

        with tqdm(
            total=total_num_generations,
            desc=f"Evolve with {self.num_islands} islands",
            unit=" generations",
            disable=not self.with_bar,
        ) as progress:

            champions_lst = []  # type: t.List[xr.Dataset]
            # Run an evolution im the archipelago several times
            for id_evolution in range(num_evolutions):
                # If the evolution on this archipelago was already run before, then
                # the migration process between the islands is automatically executed
                # Call all 'evolve()' methods on all islands
                self._pygmo_archi.evolve()
                # self._log.info(self._pygmo_archi)  # TODO: Remove this

                # Block until all evolutions have finished and raise the first exception
                # that was encountered
                self._pygmo_archi.wait_check()

                # Collect logging information from the algorithms running in the
                # archipelago
                logs.append(archi=self._pygmo_archi, id_evolution=id_evolution + 1)

                progress.update(self.algorithm.generations)

                # Get partial champions for this evolution
                partial_champions = self._get_champions()  # type: xr.Dataset

                # Get best population from the islands
                if num_best_decisions:
                    best_individuals = self.get_best_individuals(
                        num_best_decisions=num_best_decisions
                    )  # type: xr.Dataset

                    all_champions = xr.merge([partial_champions, best_individuals])
                else:
                    all_champions = partial_champions

                champions_lst.append(
                    all_champions.assign_coords(evolution=id_evolution)
                )

        champions = xr.concat(champions_lst, dim="evolution")  # type: xr.Dataset

        # Get logging information
        df_all_logs = logs.get_full_total()  # type: pd.DataFrame

        # Get the champions in a `Dataset`
        # champions = self._get_champions()  # type: # xr.Dataset
        last_champions = champions.isel(evolution=-1)

        # Get the processor(s) in a `DataFrame`
        df_results = self.problem.apply_parameters_to_processors(
            parameters=last_champions["champion_parameters"],
        )  # type: pd.DataFrame

        assert isinstance(self.problem.sim_fit_range, tuple)
        slice_times, slice_rows, slice_cols = self.problem.sim_fit_range

        geometry = self.problem.processor.detector.geometry
        no_times = len(readout.times)

        # Extract simulated 'image', 'signal' and 'pixel' from the processors
        all_simulated_full = extract_data_3d(
            df_results=df_results,
            rows=geometry.row,
            cols=geometry.col,
            times=no_times,
            readout_times=readout.times,
        )

        # Get the target data
        all_data_fit_range = all_simulated_full.sel(
            y=slice_rows, x=slice_cols, readout_time=slice_times
        )
        if readout.time_domain_simulation:
            all_data_fit_range["target"] = xr.DataArray(
                self.problem.all_target_data,
                dims=["id_processor", "readout_time", "y", "x"],
            )
        else:
            all_data_fit_range["target"] = xr.DataArray(
                self.problem.all_target_data, dims=["id_processor", "y", "x"]
            )

        ds = xr.merge([champions, all_data_fit_range])  # type: xr.Dataset

        ds.attrs["num_islands"] = self.num_islands
        ds.attrs["population_size"] = self.algorithm.population_size
        ds.attrs["num_evolutions"] = num_evolutions
        ds.attrs["generations"] = self.algorithm.generations

        ds = ds.assign_coords({"param_id": range(ds.dims["param_id"])})

        return ds, df_results, df_all_logs

        # # Get the processor(s) in a `DataFrame`
        # df_processors = self.problem.apply_parameters_to_processors(
        #     parameters=last_champions["champion_parameters"],
        # )  # type: pd.DataFrame
        #
        # assert isinstance(self.problem.sim_fit_range, tuple)
        # slice_rows, slice_cols = self.problem.sim_fit_range
        #
        # geometry = self.problem.processor.detector.geometry
        #
        # # Extract simulated 'image', 'signal' and 'pixel' from the processors
        # all_simulated_full = extract_data_2d(
        #     df_processors=df_processors,
        #     rows=geometry.row,
        #     cols=geometry.col,
        # )
        #
        # # Get the target data
        # all_data_fit_range = all_simulated_full.sel(y=slice_rows, x=slice_cols)
        # all_data_fit_range["target"] = xr.DataArray(
        #     self.problem.all_target_data, dims=["id_processor", "y", "x"]
        # )
        #
        # ds = xr.merge([champions, all_data_fit_range])  # type: xr.Dataset
        #
        # ds.attrs["num_islands"] = self.num_islands
        # ds.attrs["population_size"] = self.algorithm.population_size
        # ds.attrs["num_evolutions"] = num_evolutions
        # ds.attrs["generations"] = self.algorithm.generations
        #
        # ds = ds.assign_coords({"param_id": range(ds.dims["param_id"])})
        #
        # return ds, df_processors, df_all_logs

    def _get_champions(self) -> xr.Dataset:
        """Extract the champions.

        Returns
        -------
        Dataset
            A dataset containing the champions.

        Examples
        --------
        >>> self._get_champions()
        <xarray.Dataset>
        Dimensions:              (island: 2, param_id: 7)
        Dimensions without coordinates: island, param_id
        Data variables:
            champion_fitness     (island) float64 3.285e+04 4.102e+04
            champion_decision    (island, param_id) float64 0.1526 -1.977 ... 0.9329
            champion_parameters  (island, param_id) float64 0.1526 -1.977 ... 8.568
        """
        # Get fitness and decision vectors of the num_islands' champions
        champions_1d_fitness = self._pygmo_archi.get_champions_f()  # type: ArrayLike

        champions_1d_decision = self._pygmo_archi.get_champions_x()  # type: ArrayLike

        # Get the champions as a Dataset
        champions = xr.Dataset()
        champions["champion_fitness"] = xr.DataArray(
            np.ravel(champions_1d_fitness), dims="island"
        )
        champions["champion_decision"] = xr.DataArray(
            champions_1d_decision, dims=["island", "param_id"]
        )
        champions["champion_parameters"] = xr.DataArray(
            self.problem.convert_to_parameters(champions["champion_decision"]),
            dims=["island", "param_id"],
        )

        return champions

    def get_best_individuals(self, num_best_decisions: int) -> xr.Dataset:
        """Get the best decision vectors and fitness from the island of an archipelago.

        Parameters
        ----------
        num_best_decisions : int or None, optional.
            Number of best individuals to extract. If this parameter is set to None then
            no individuals are extracted.

        Returns
        -------
        Dataset
            A new dataset with two data arrays 'best_decision' and 'best_fitness'.

        Examples
        --------
        >>> archi = MyArchipelago(...)
        >>> archi.get_best_individuals(num_best_decisions=5)
        <xarray.Dataset>
        Dimensions:          (individual: 10, island: 2, param_id: 7)
        Coordinates:
          * island           (island) int64 0 1
          * individual       (individual) int64 0 1 2 3 4 5 6 7 8 9
        Dimensions without coordinates: param_id
        Data variables:
            best_decision    (island, individual, param_id) float64 0.1526 ... 0.1608
            best_parameters  (island, individual, param_id) float64 0.1526 ... 0.1608
            best_fitness     (island, individual) float64 3.285e+04 ... 5.732e+04

        Raises
        ------
        ValueError
            Raised if 'num_best_decisions' is a negative 'int' value.
        """
        if num_best_decisions < 0:
            raise ValueError(
                "'num_best_decisions' must be 'None' or a positive integer"
            )

        lst = []
        for island_idx, island in enumerate(self._pygmo_archi):
            population = island.get_population()  # type: pg.population

            # Get the decision vectors: num_individuals x size_decision_vector
            decision_vectors_2d = population.get_x()  # type: np.ndarray

            # Get the fitness vectors: num_individuals x 1
            fitness_vectors_2d = population.get_f()  # type: np.ndarray

            # Convert the decision vectors to parameters:
            #   num_individuals x size_decision_vector
            parameters_2d = self.problem.convert_to_parameters(decision_vectors_2d)

            # Add the vectors into an Dataset
            island_population = xr.Dataset()
            island_population["best_decision"] = xr.DataArray(
                decision_vectors_2d, dims=["individual", "param_id"]
            )
            island_population["best_parameters"] = xr.DataArray(
                parameters_2d, dims=["individual", "param_id"]
            )
            island_population["best_fitness"] = xr.DataArray(
                fitness_vectors_2d.flatten(), dims=["individual"]
            )

            # Get the indexes for the best fitness vectors
            # and extract the 'num_besnum_best_decisionst_decisions' individuals
            all_indexes_sorted = island_population["best_fitness"].argsort()
            first_indexes_sorted = all_indexes_sorted[:num_best_decisions]

            # Use the indexes to get the best elements
            island_best_population = island_population.sel(
                individual=first_indexes_sorted
            )

            # Append the result and add a new coordinate 'island'
            lst.append(island_best_population.assign_coords(island=island_idx))

        # Create a new dataset
        best_individuals_no_coordinates = xr.concat(lst, dim="island")

        # Add coordinates
        num_individuals = len(best_individuals_no_coordinates["individual"])
        best_individuals = best_individuals_no_coordinates.assign_coords(
            individual=range(num_individuals),
            island=range(len(self._pygmo_archi)),
        )  # type: xr.Dataset

        return best_individuals



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Readout noise model."""
import logging
import typing as t

import numpy as np

from pyxel.detectors import CMOS, Detector

# from astropy import units as u


# @validators.validate
# @config.argument(name='', label='', units='', validate=)


def output_node_noise(
    detector: Detector, std_deviation: float, random_seed: t.Optional[int] = None
) -> None:
    """Adding noise to signal array of detector output node using normal random distribution.

    detector Signal unit: Volt

    :param detector: Pyxel Detector object
    :param std_deviation: standard deviation
    :param random_seed: seed
    """
    logging.info("")
    if random_seed:
        np.random.seed(random_seed)

    signal_mean_array = detector.signal.array.astype("float64")
    sigma_array = std_deviation * np.ones(signal_mean_array.shape)

    signal = np.random.normal(loc=signal_mean_array, scale=sigma_array)

    detector.signal.array = signal


def output_node_noise_cmos(
    detector: "CMOS",
    readout_noise: float,
    readout_noise_std: float,
    random_seed: t.Optional[int] = None,
) -> None:
    """Output node noise model for CMOS detectors where readout is statistically independent for each pixel.

    Parameters
    ----------
    detector
    readout_noise: Mean readout noise for the array in units of electrons.
    readout_noise_std: Readout noise standard deviation in units of electrons.
    random_seed

    Returns
    -------
    None
    """
    logging.info("")
    if random_seed:
        np.random.seed(random_seed)

    # sv is charge readout sensitivity
    sv = detector.characteristics.sv

    signal_mean_array = detector.signal.array.astype("float64")
    sigma_array = np.random.normal(
        loc=readout_noise * sv,
        scale=readout_noise_std * sv,
        size=signal_mean_array.shape,
    )

    signal = np.random.normal(loc=signal_mean_array, scale=sigma_array)

    detector.signal.array = signal



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Functions for assigning custom QE maps."""

import typing as t
from pathlib import Path

from pyxel import load_image

if t.TYPE_CHECKING:
    from pyxel.detectors import Detector


def qe_map(
    detector: "Detector",
    filename: t.Union[str, Path],
    fit_qe_to_det: bool = False,
    position_x: int = 0,
    position_y: int = 0,
) -> None:
    """Upload from file and assign a custom QE map to a detector.

    Parameters
    ----------
    detector: Detector
    filename: str or Path
    fit_qe_to_det: bool
    position_x: int
    position_y: int

    Returns
    -------
    None
    """
    qe = load_image(filename)

    rows = detector.geometry.row
    columns = detector.geometry.col

    if rows + position_y > qe.shape[0]:
        raise ValueError("Shapes do not match in Y direction.")
    if columns + position_x > qe.shape[1]:
        raise ValueError("Shapes do not match in X direction.")

    if fit_qe_to_det:
        qe = qe[
            slice(position_y, position_y + rows),
            slice(position_x, position_x + columns),
        ]

    detector.characteristics.qe = qe



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Charge Distortion Model for CCDs.

============================
This is a function to run the upgraded CDM CTI model developed by Alex Short (ESA).

:requires: NumPy

:author: Alex Short
:author: Sami-Matias Niemi
:author: David Lucsanyi
"""
import logging
import typing as t

import numba
import numpy as np

from pyxel.detectors import CCD

try:
    from matplotlib import pyplot as plt
except ImportError:
    # raise Warning('Matplotlib cannot be imported')
    pass


# @pyxel.validate
# @pyxel.argument(name='', label='', units='', validate=)
# @pyxel.register(group='charge_transfer', name='cdm', detector='ccd')
# TODO: What is the type of 'tr_p', 'tr_s', 'nt_p', 'nt_s', 'sigma_p' and 'sigma_s' ?
#       Is it `t.Union[float, t.Iterable[float]]` ???
def cdm(
    detector: CCD,
    parallel_cti: bool,
    serial_cti: bool,
    beta_p: float,
    beta_s: float,
    tr_p: t.Sequence[float],
    tr_s: t.Sequence[float],
    nt_p: t.Sequence[float],
    nt_s: t.Sequence[float],
    sigma_p: t.Sequence[float],
    sigma_s: t.Sequence[float],
    charge_injection: bool,
) -> None:
    """Charge Distortion Model (CDM) model wrapper.

    :param detector: Pyxel CCD detector object
    :param parallel_cti: switch on CTI in parallel direction (along column)
    :param serial_cti: switch on CTI in serial direction (along rows)
    :param charge_injection: set this true in case of charge injection,
        charge packets goes through all pixels in parallel direction
    :param beta_p: electron cloud expansion coefficient, parallel
    :param beta_s: electron cloud expansion coefficient, serial
    :param tr_p: trap release time constants (τ_r), parallel
    :param tr_s: trap release time constants (τ_r), serial
    :param nt_p: absolute trap densities (n_t), parallel
    :param nt_s: absolute trap densities (n_t), serial
    :param sigma_p: trap capture cross-sections (σ), parallel
    :param sigma_s: trap capture cross-sections (σ), serial
    """
    # Ne - number of electrons in a pixel
    # ne - electron density in the vicinity of the trap
    # Vc - volume of the charge cloud
    # Pr - the probability that the trap will release the electron into the sample
    # tau_c - capture time constant
    # Pc - capture probability (per vacant trap) as a function of the number of sample electrons Ne
    # NT - number of traps in the column,
    # NT = 2*nt*Vg*x  where x is the number of TDI transfers or the column length in pixels.
    # Nc - number of electrons captured by a given trap species during the transit of an integrating signal packet
    # N0 - initial trap occupancy
    # Nr - number of electrons released into the sample during a transit along the column
    # vg: assumed maximum geometrical volume electrons can occupy within a pixel (parallel)
    # svg: assumed maximum geometrical volume electrons can occupy within a pixel (serial)
    # t: constant TDI period (parallel)
    # st: constant TDI period (serial)

    logging.info("")
    char = detector.characteristics

    detector.pixel.array = run_cdm(
        s=detector.pixel.array,
        vg=char.vg,
        svg=char.svg,
        t=char.t,
        st=char.st,
        fwc=char.fwc,
        sfwc=char.fwc_serial,
        vth=detector.e_thermal_velocity,
        parallel_cti=parallel_cti,
        serial_cti=serial_cti,
        charge_injection=charge_injection,
        chg_inj_parallel_transfers=detector.geometry.row,
        beta_p=beta_p,
        beta_s=beta_s,
        tr_p=np.array(tr_p),
        tr_s=np.array(tr_s),
        nt_p=np.array(nt_p),
        nt_s=np.array(nt_s),
        sigma_p=np.array(sigma_p),
        sigma_s=np.array(sigma_s),
    )


@numba.njit(nogil=True)
def run_cdm(
    s: np.ndarray,
    beta_p: float,
    beta_s: float,
    vg: float,
    svg: float,
    t: float,
    st: float,
    fwc: float,
    sfwc: float,
    vth: float,
    tr_p: np.ndarray,
    tr_s: np.ndarray,
    nt_p: np.ndarray,
    nt_s: np.ndarray,
    sigma_p: np.ndarray,
    sigma_s: np.ndarray,
    charge_injection: bool = False,
    chg_inj_parallel_transfers: int = 0,
    parallel_cti: bool = True,
    serial_cti: bool = True,
) -> np.ndarray:
    """CDM model.

    :param s: np.ndarray
    :param dob:
    :param beta_p: electron cloud expansion coefficient (parallel)
    :param beta_s: electron cloud expansion coefficient (serial)
    :param vg: assumed maximum geometrical volume electrons can occupy within a pixel (parallel)
    :param svg: assumed maximum geometrical volume electrons can occupy within a pixel (serial)
    :param t: constant TDI period (parallel)
    :param st: constant TDI period (serial)
    :param fwc:
    :param sfwc:
    :param vth:
    :param charge_injection:
    :param chg_inj_parallel_transfers:
    :param sigma_p:
    :param sigma_s:
    :param tr_p:
    :param tr_s:
    :param nt_p: number of traps per electron cloud (and not pixel!) in parallel direction
    :param nt_s: number of traps per electron cloud (and not pixel!) in serial direction
    :param parallel_cti:
    :param serial_cti:
    :return:
    """
    ydim, xdim = s.shape  # full signal array we want to apply cdm for

    kdim_p = len(nt_p)
    kdim_s = len(nt_s)

    # np.clip(s, 0., fwc, s)      # full well capacity

    nt_p = nt_p / vg  # parallel trap density (traps / cm**3)
    nt_s = nt_s / svg  # serial trap density (traps / cm**3)

    # nt_p *= rdose             # absolute trap density [per cm**3]
    # nt_s *= rdose             # absolute trap density [per cm**3]

    # IMAGING (non-TDI) MODE
    # Parallel direction
    if parallel_cti:
        no = np.zeros((xdim, kdim_p))
        alpha_p = t * sigma_p * vth * fwc ** beta_p / (2.0 * vg)  # type: np.ndarray
        g_p = 2.0 * nt_p * vg / fwc ** beta_p  # type: np.ndarray
        for i in range(0, ydim):
            if charge_injection:
                gamma_p = (
                    g_p * chg_inj_parallel_transfers
                )  # number of all transfers in parallel dir.
            else:
                gamma_p = g_p * i  # TODO: (i+1) ?????????????
            for k in range(kdim_p):
                for j in range(xdim):
                    nc = 0.0
                    if s[i, j] > 0.01:
                        nc = max(
                            (gamma_p[k] * s[i, j] ** beta_p - no[j, k])
                            / (gamma_p[k] * s[i, j] ** (beta_p - 1.0) + 1.0)
                            * (
                                1.0
                                - np.exp(-1 * alpha_p[k] * s[i, j] ** (1.0 - beta_p))
                            ),
                            0.0,
                        )
                        no[j, k] += nc
                    nr = no[j, k] * (1.0 - np.exp(-t / tr_p[k]))
                    s[i, j] += -1 * nc + nr
                    no[j, k] -= nr
                    if s[i, j] < 0.01:
                        s[i, j] = 0.0

    # IMAGING (non-TDI) MODE
    # Serial direction
    if serial_cti:
        sno = np.zeros((ydim, kdim_s))
        alpha_s = st * sigma_s * vth * sfwc ** beta_s / (2.0 * svg)  # type: np.ndarray
        g_s = 2.0 * nt_s * svg / sfwc ** beta_s
        for j in range(0, xdim):
            gamma_s = g_s * j  # TODO: (j+1) ?????????????
            for k in range(kdim_s):
                for i in range(ydim):
                    nc = 0.0
                    if s[i, j] > 0.01:
                        nc = max(
                            (gamma_s[k] * s[i, j] ** beta_s - sno[i, k])
                            / (gamma_s[k] * s[i, j] ** (beta_s - 1.0) + 1.0)
                            * (
                                1.0
                                - np.exp(-1 * alpha_s[k] * s[i, j] ** (1.0 - beta_s))
                            ),
                            0.0,
                        )
                        sno[i, k] += nc
                    nr = sno[i, k] * (1.0 - np.exp(-st / tr_s[k]))
                    s[i, j] += -1 * nc + nr
                    sno[i, k] -= nr
                    if s[i, j] < 0.01:
                        s[i, j] = 0.0

    return s


def plot_serial_profile(
    data: np.ndarray, row: int, data2: t.Optional[np.ndarray] = None
) -> None:
    """TBW.

    :param data:
    :param row:
    :param data2:
    """
    ydim, xdim = data.shape
    profile_x = list(range(ydim))
    profile_y = data[row, :]
    plt.title("Serial profile")
    plt.plot(profile_x, profile_y, color="blue")
    if data2 is not None:
        profile_y_2 = data2[row, :]
        plt.plot(profile_x, profile_y_2, color="red")


def plot_parallel_profile(
    data: np.ndarray, col: int, data2: t.Optional[np.ndarray] = None
) -> None:
    """TBW.

    :param data:
    :param col:
    :param data2:
    """
    ydim, xdim = data.shape
    profile_x = list(range(xdim))
    profile_y = data[:, col]
    plt.title("Parallel profile")
    plt.plot(profile_x, profile_y, color="blue")
    if data2 is not None:
        profile_y_2 = data2[:, col]
        plt.plot(profile_x, profile_y_2, color="red")


def plot_1d_profile(
    array: np.ndarray, offset: int = 0, label: str = "", m: str = "-"
) -> None:
    """Plot profile on log scale.

    :param array:
    :param offset:
    :param label:
    :param m:
    """
    x = list(range(offset, offset + len(array)))
    # plt.title('Parallel profile, charge injection')
    plt.semilogy(x, array, m, label=label)
    if label:
        plt.legend()


def plot_1d_profile_lin(
    array: np.ndarray,
    offset: int = 0,
    label: str = "",
    m: str = "-",
    col: t.Optional[str] = None,
) -> None:
    """TBW.

    :param array:
    :param offset:
    :param label:
    :param m:
    :param col:
    """
    x = list(range(offset, offset + len(array)))
    # plt.title('Parallel profile, charge injection')
    plt.plot(x, array, m, label=label, color=col)
    if label:
        plt.legend()


def plot_1d_profile_with_err(
    array: np.ndarray, error: np.ndarray, offset: int = 0, label: str = ""
) -> None:
    """TBW.

    :param array:
    :param error:
    :param offset:
    :param label:
    """
    x = list(range(offset, offset + len(array)))
    plt.title("Parallel profile with error, charge injection")
    plt.errorbar(x, array, error, label=label, linestyle="None", marker=".")
    if label:
        plt.legend()


def plot_residuals(
    data: np.ndarray, data2: np.ndarray, label: str = ""
) -> None:  # col='magenta',
    """TBW.

    :param data:
    :param data2:
    # :param col:
    :param label:
    """
    x = list(range(len(data)))
    # plt.title('Residuals of fitted and target parallel CTI profiles')
    residuals = np.around(data - data2, decimals=5)
    # residuals = data-data2
    # plt.plot(x, residuals, '.', color=col, label=label)
    plt.plot(x, residuals, ".", label=label)
    plt.legend()


def plot_image(data: np.ndarray) -> None:
    """TBW.

    :param data:
    """
    plt.imshow(data, cmap=plt.gray())  # , interpolation='nearest')
    plt.xlabel("x - serial direction")
    plt.ylabel("y - parallel direction")
    plt.title("CCD image with CTI")
    plt.colorbar()


# @numba.jit(nopython=True, nogil=True, parallel=True)
def optimized_cdm(
    s: np.ndarray,
    beta_p: float,
    beta_s: float,
    vg: float,
    svg: float,
    t: float,
    st: float,
    fwc: float,
    sfwc: float,
    vth: float,
    tr_p: np.ndarray,
    tr_s: np.ndarray,
    nt_p: np.ndarray,
    nt_s: np.ndarray,
    sigma_p: np.ndarray,
    sigma_s: np.ndarray,
    charge_injection: bool = False,
    chg_inj_parallel_transfers: int = 0,
    parallel_cti: bool = True,
    serial_cti: bool = True,
) -> np.ndarray:
    """CDM model.

    Done by Patricia Liebing. Not yet test or compared to the original 'run_cdm' function.

    :param s: np.ndarray
    :param dob:
    :param beta_p: electron cloud expansion coefficient (parallel)
    :param beta_s: electron cloud expansion coefficient (serial)
    :param vg: assumed maximum geometrical volume electrons can occupy within
               a pixel (parallel)
    :param svg: assumed maximum geometrical volume electrons can occupy within
                a pixel (serial)
    :param t: constant TDI period (parallel)
    :param st: constant TDI period (serial)
    :param fwc:
    :param sfwc:
    :param vth:
    :param charge_injection:
    :param chg_inj_parallel_transfers:
    :param sigma_p:
    :param sigma_s:
    :param tr_p:
    :param tr_s:
    :param nt_p: number of traps per electron cloud (and not pixel!)
                 in parallel direction
    :param nt_s: number of traps per electron cloud (and not pixel!)
                 in serial direction
    :param parallel_cti:
    :param serial_cti:
    :return:
    """
    ydim, xdim = s.shape  # full signal array we want to apply cdm for

    kdim_p = len(nt_p)
    kdim_s = len(nt_s)

    # np.clip(s, 0., fwc, s)      # full well capacity

    nt_p = nt_p / vg  # parallel trap density (traps / cm**3)
    nt_s = nt_s / svg  # serial trap density (traps / cm**3)

    # nt_p *= rdose             # absolute trap density [per cm**3]
    # nt_s *= rdose             # absolute trap density [per cm**3]

    # IMAGING (non-TDI) MODE
    # Parallel direction
    if parallel_cti:
        no = np.zeros((xdim, kdim_p))
        alpha_p = t * sigma_p * vth * fwc ** beta_p / (2.0 * vg)  # type: np.ndarray
        g_p = 2.0 * nt_p * vg / fwc ** beta_p  # type: np.ndarray
        for i in range(0, ydim):
            if charge_injection:
                gamma_p = (
                    g_p * chg_inj_parallel_transfers
                )  # number of all transfers in parallel dir.
            else:
                gamma_p = g_p * (i + 1)
            for k in range(kdim_p):
                nc = np.zeros(xdim)
                nctest = np.zeros(xdim)
                sclm = s[i, :]
                idx = np.where(sclm > 0.01)
                nok = no[:, k]
                nc[idx] += (
                    (gamma_p[k] * sclm[idx] ** beta_p - nok[idx])
                    / (gamma_p[k] * sclm[idx] ** (beta_p - 1.0) + 1.0)
                    * (1.0 - np.exp(-1 * alpha_p[k] * sclm[idx] ** (1.0 - beta_p)))
                )
                no[:, k] += np.fmax(nctest, nc)
                nr = no[:, k] * (1.0 - np.exp(-t / tr_p[k]))
                s[i, :] += -1 * nc + nr
                no[:, k] -= nr

    # IMAGING (non-TDI) MODE
    # Serial direction
    if serial_cti:
        sno = np.zeros((ydim, kdim_s))
        alpha_s = st * sigma_s * vth * sfwc ** beta_s / (2.0 * svg)  # type: np.ndarray
        g_s = 2.0 * nt_s * svg / sfwc ** beta_s  # type: np.ndarray
        for j in range(0, xdim):
            gamma_s = g_s * (j + 1)
            for k in range(kdim_s):
                ncs = np.zeros(ydim)
                ncstest = np.zeros(ydim)
                srow = s[:, j]
                idxs = np.where(srow > 0.01)
                snok = sno[:, k]
                ncs[idxs] += (
                    (gamma_s[k] * srow[idxs] ** beta_s - snok[idxs])
                    / (gamma_s[k] * srow[idxs] ** (beta_s - 1.0) + 1.0)
                    * (1.0 - np.exp(-1 * alpha_s[k] * srow[idxs] ** (1.0 - beta_s)))
                )
                sno[:, k] += np.fmax(ncstest, ncs)
                nrs = sno[:, k] * (1.0 - np.exp(-st / tr_s[k]))
                s[:, j] += -1 * ncs + nrs
                sno[:, k] -= nrs

    return s



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Object memory consumption utilities."""

import typing as t
from typing import List

import numpy as np
from pympler.asizeof import asizeof  # type: ignore


def get_size(obj: t.Any) -> int:
    """Recursively calculates object size in bytes using Pympler library.

    Parameters
    ----------
    obj: object
        Any input object.

    Returns
    -------
    int
        Object size in bytes.
    """
    return int(asizeof(obj))


def print_human_readable_memory(usage: dict) -> None:
    """Convert byte sizes from a dictionary and print a human readable form.

    Parameters
    ----------
    usage: dict

    Returns
    -------
    None
    """
    for k, v in usage.items():
        for unit in ["Bytes", "KB", "MB", "GB"]:
            if v < 1024.0:
                print(f"{k:<20}{np.round(v,decimals=1):<7}{unit}")
                break
            v /= 1024.0


def memory_usage_details(
    obj: t.Any,
    *attr_kw: List[str],
    print_result: bool = True,
    human_readable: bool = True,
) -> dict:
    """TBW.

    Parameters
    ----------
    obj: object
    attr_kw: list
    print_result: bool
    human_readable: bool

    Returns
    -------
    usage: dict

    Raises
    ------
    Value Error
        When no attributes provided.
    AttributeError
        Attribute does not exist in the object.
    """

    usage: dict = {}

    if attr_kw == ():
        raise ValueError("No attributes provided.")

    for attribute in attr_kw[0]:
        if attribute not in obj.__dict__.keys():
            raise AttributeError(
                f"Attribute {attribute} not found in the observed object."
            )

    for key, value in obj.__dict__.items():
        if hasattr(value, "numbytes") and key in attr_kw[0]:
            usage.update({key.replace("_", ""): get_size(value)})

    if print_result:
        if human_readable:
            print_human_readable_memory(usage)
        else:
            print(usage)

    return usage



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""CDM model calibration with PYGMO.

https://esa.github.io/pagmo2/index.html
"""
import copy
import logging
import math
import typing as t
from copy import deepcopy
from numbers import Number
from pathlib import Path

import dask.delayed as delayed
import numpy as np
import pandas as pd
from typing_extensions import Literal

from pyxel.calibration import (
    CalibrationMode,
    ProblemSingleObjective,
    check_ranges,
    list_to_3d_slice,
    list_to_slice,
    read_data,
    read_datacubes,
)
from pyxel.exposure import run_exposure_pipeline
from pyxel.observation.parameter_values import ParameterValues
from pyxel.pipelines import Processor
from pyxel.pipelines.processor import ResultType

if t.TYPE_CHECKING:
    import xarray as xr
    from numpy.typing import ArrayLike

    from pyxel.exposure import Readout


class ModelFitting(ProblemSingleObjective):
    """Pygmo problem class to fit data with any model in Pyxel."""

    def __init__(
        self,
        processor: Processor,
        variables: t.Sequence[ParameterValues],
        readout: "Readout",
    ):
        self.processor = processor  # type: Processor
        self.variables = variables  # type: t.Sequence[ParameterValues]

        self.calibration_mode = None  # type: t.Optional[CalibrationMode]
        self.original_processor = None  # type: t.Optional[Processor]
        self.generations = None  # type: t.Optional[int]
        self.pop = None  # type: t.Optional[int]
        self.readout = readout  # type: Readout

        self.all_target_data = []  # type: t.List[np.ndarray]
        self.weighting = None  # type: t.Optional[np.ndarray]
        self.weighting_from_file = None  # type: t.Optional[t.Sequence[np.ndarray]]
        self.fitness_func = None  # type: t.Optional[t.Callable]
        self.sim_output = None  # type: t.Optional[ResultType]
        # self.fitted_model = None            # type: t.Optional['ModelFunction']
        self.param_processor_list = []  # type: t.List[Processor]

        self.file_path = None  # type: t.Optional[Path]

        self.fitness_array = None  # type: t.Optional[np.ndarray]
        self.population = None  # type: t.Optional[np.ndarray]
        self.champion_f_list = None  # type: t.Optional[np.ndarray]
        self.champion_x_list = None  # type: t.Optional[np.ndarray]

        self.lbd = []  # type: t.Sequence[float]  # lower boundary
        self.ubd = []  # type: t.Sequence[float]  # upper boundary

        self.sim_fit_range = (
            slice(None),
            slice(None),
            slice(None),
        )  # type: t.Tuple[slice, slice, slice]
        self.targ_fit_range = (
            slice(None),
            slice(None),
        )  # type: t.Union[t.Tuple[slice, slice], t.Tuple[slice, slice, slice]]

        self.match = {}  # type: t.Dict[int, t.List[str]]

        # self.normalization = False
        # self.target_data_norm = []

    def get_bounds(self) -> t.Tuple[t.Sequence[float], t.Sequence[float]]:
        """Get the box bounds of the problem (lower_boundary, upper_boundary).

        It also implicitly defines the dimension of the problem.

        Returns
        -------
        tuple of lower boundaries and upper boundaries
        """
        return self.lbd, self.ubd

    def configure(
        self,
        calibration_mode: CalibrationMode,
        simulation_output: ResultType,
        fitness_func: t.Callable,
        population_size: int,
        generations: int,
        file_path: t.Optional[Path],
        target_fit_range: t.Sequence[int],
        out_fit_range: t.Sequence[int],
        target_output: t.Sequence[Path],
        input_arguments: t.Optional[t.Sequence[ParameterValues]] = None,
        weights: t.Optional[t.Sequence[float]] = None,
        weights_from_file: t.Optional[t.Sequence[Path]] = None,
    ) -> None:
        """TBW.

        Parameters
        ----------
        sampling
        calibration_mode
        simulation_output
        fitness_func
        population_size
        generations
        file_path
        target_fit_range
        out_fit_range
        target_output
        input_arguments
        weights
        weights_from_file
        """
        self.calibration_mode = CalibrationMode(calibration_mode)
        self.sim_output = ResultType(simulation_output)
        self.fitness_func = fitness_func
        self.pop = population_size
        self.generations = generations

        # TODO: Remove 'assert'
        # assert isinstance(self.pop, int)

        # if self.calibration_mode == 'single_model':           # TODO update
        #     self.single_model_calibration()

        self.set_bound()

        self.original_processor = deepcopy(self.processor)
        if input_arguments:

            max_val, min_val = 0, 1000
            for arg in input_arguments:
                min_val = min(min_val, len(arg.values))
                max_val = max(max_val, len(arg.values))
            if min_val != max_val:
                logging.warning(
                    'The "result_input_arguments" value lists have different lengths! '
                    "Some values will be ignored."
                )
            for i in range(min_val):
                new_processor = deepcopy(self.processor)  # type: Processor
                for step in input_arguments:  # type: ParameterValues
                    assert step.values != "_"

                    value = step.values[i]  # type: t.Union[Literal['_'], str, Number]

                    step.current = value
                    new_processor.set(key=step.key, value=step.current)
                self.param_processor_list += [new_processor]
        else:
            self.param_processor_list = [deepcopy(self.processor)]

        params = 0  # type: int
        for var in self.variables:  # type: ParameterValues
            if isinstance(var.values, list):
                b = len(var.values)
            else:
                b = 1

            params += b
        self.champion_f_list = np.zeros((1, 1))
        self.champion_x_list = np.zeros((1, params))
        self.file_path = file_path

        if self.readout.time_domain_simulation:
            target_list_3d = read_datacubes(
                filenames=target_output
            )  # type: t.Sequence[np.ndarray]
            times, rows, cols = target_list_3d[0].shape
            check_ranges(
                target_fit_range=target_fit_range,
                out_fit_range=out_fit_range,
                rows=rows,
                cols=cols,
                readout_times=times,
            )
            self.targ_fit_range = list_to_slice(target_fit_range)
            self.sim_fit_range = list_to_3d_slice(out_fit_range)
            for target_3d in target_list_3d:  # type: np.ndarray
                self.all_target_data += [target_3d[self.targ_fit_range]]

        else:
            target_list_2d = read_data(
                filenames=target_output
            )  # type: t.Sequence[np.ndarray]
            rows, cols = target_list_2d[0].shape
            check_ranges(
                target_fit_range=target_fit_range,
                out_fit_range=out_fit_range,
                rows=rows,
                cols=cols,
            )
            self.targ_fit_range = list_to_slice(target_fit_range)
            out_fit_range = [None, None] + out_fit_range  # type: ignore
            self.sim_fit_range = list_to_3d_slice(out_fit_range)
            for target_2d in target_list_2d:  # type: np.ndarray
                self.all_target_data += [target_2d[self.targ_fit_range]]

            self._configure_weights(
                weights=weights, weights_from_file=weights_from_file
            )

    def _configure_weights(
        self,
        weights: t.Optional[t.Sequence[float]] = None,
        weights_from_file: t.Optional[t.Sequence[Path]] = None,
    ) -> None:
        """TBW.

        Parameters
        ----------
        weights
        weights_from_file

        Returns
        -------
        None
        """
        if weights_from_file is not None:
            if self.readout.time_domain_simulation:
                wf = read_datacubes(weights_from_file)
                self.weighting_from_file = [
                    weight_array[self.targ_fit_range] for weight_array in wf
                ]
            else:
                wf = read_data(weights_from_file)
                self.weighting_from_file = [
                    weight_array[self.targ_fit_range] for weight_array in wf
                ]
        elif weights is not None:
            self.weighting = np.array(weights)

    # def single_model_calibration(self):     # TODO update
    #     """TBW.
    #
    #     :return:
    #     """
    #     # if len(self.model_name_list) > 1:
    #     #     raise ValueError('Select only one pipeline model!')
    #     # if self.model_name_list[0] in ['geometry', 'material', 'environment', 'characteristics']:
    #     #     raise ValueError('Select a pipeline model and not a detector attribute!')
    #
    #     self.fitted_model = self.processor.pipeline.get_model(self.model_name_list[0])
    #     self.processor.run_pipeline(abort_before=self.model_name_list[0])

    def set_bound(self) -> None:
        """TBW."""
        self.lbd = []
        self.ubd = []
        for var in self.variables:  # type: ParameterValues
            assert var.boundaries
            low_val, high_val = var.boundaries  # type: t.Tuple[float, float]

            if var.logarithmic:
                low_val = math.log10(low_val)
                high_val = math.log10(high_val)

            if var.values == "_":
                self.lbd += [low_val]
                self.ubd += [high_val]
            elif isinstance(var.values, list) and all(x == "_" for x in var.values[:]):
                self.lbd += [low_val] * len(var.values)
                self.ubd += [high_val] * len(var.values)
            else:
                raise ValueError(
                    'Character "_" (or a list of it) should be used to '
                    "indicate variables need to be calibrated"
                )

    def get_simulated_data(self, processor: Processor) -> np.ndarray:
        """Extract 2D data from a processor."""
        if self.sim_output == ResultType.Image:
            simulated_data = processor.result["image"][
                self.sim_fit_range
            ]  # type: np.ndarray

        elif self.sim_output == ResultType.Signal:
            simulated_data = processor.result["signal"][self.sim_fit_range]
        elif self.sim_output == ResultType.Pixel:
            simulated_data = processor.result["pixel"][self.sim_fit_range]

        return simulated_data

    # def batch_fitness(self, population_parameter_vector: np.ndarray) -> np.ndarray:
    #     """Batch Fitness Evaluation.
    #
    #     PYGMO BFE IS STILL NOT FULLY IMPLEMENTED, THEREFORE THIS FUNC CAN NOT BE USED YET.
    #
    #     Parameters
    #     ----------
    #     population_parameter_vector : array-like
    #         The decision vectors (chromosomes) to be evaluated in batch mode
    #
    #     Returns
    #     -------
    #     array : 1D numpy float array
    #         The fitness vectors of input 'population_parameter_vector'.
    #     """
    #     logger = logging.getLogger("pyxel")
    #     logger.info("batch_fitness() called with %s " % population_parameter_vector)
    #
    #     fitness_vector = []  # type: t.List[float]
    #     for parameter in population_parameter_vector:
    #         overall_fitness = 0.0
    #         parameter = self.old_update_parameter([parameter])
    #         processor_list = deepcopy(self.param_processor_list)
    #
    #         for processor, target_data in zip(processor_list, self.all_target_data):
    #             processor = self.old_update_processor(parameter, processor)
    #             # processor = delayed(self.old_update_processor)(parameter, processor)
    #
    #             result_proc = processor.run_pipeline()
    #             # result_proc = delayed(processor.run_pipeline)()
    #
    #             simulated_data = self.get_simulated_data(result_proc)
    #             # simulated_data = delayed(self.get_simulated_data)(result_proc)
    #
    #             fitness = self.calculate_fitness(simulated_data, target_data)
    #             # fitness = delayed(self.calculate_fitness)(simulated_data, target_data)
    #
    #             overall_fitness = add(overall_fitness, fitness)
    #             # overall_fitness = delayed(add)(overall_fitness, fitness)
    #
    #         fitness_vector.append(
    #             overall_fitness
    #         )  # overall fitness per individual for the full population
    #
    #     fitness_vector = merge_fitness(fitness_vector)
    #     # fitness_vector_delayed = delayed(merge_fitness)(fitness_vector)  # type: Delayed
    #
    #     population_fitness_vector = fitness_vector
    #     # population_fitness_vector = fitness_vector_delayed.compute()
    #
    #     return population_fitness_vector

    def calculate_fitness(
        self,
        simulated_data: np.ndarray,
        target_data: np.ndarray,
        weighting: t.Optional[np.ndarray] = None,
    ) -> float:
        """TBW.

        Parameters
        ----------
        simulated_data
        target_data
        weighting
        """
        # TODO: Remove 'assert'
        assert self.fitness_func is not None

        if weighting is not None:
            factor = weighting
        else:
            factor = np.ones(np.shape(target_data))

        fitness = self.fitness_func(
            simulated=simulated_data.astype(np.float64),
            target=target_data.astype(np.float64),
            weighting=factor.astype(np.float64),
        )  # type: float

        return fitness

    # TODO: If possible, use 'numba' for this method
    def fitness(self, decision_vector_1d: np.ndarray) -> t.Sequence[float]:
        """Call the fitness function, elements of parameter array could be logarithmic values.

        Parameters
        ----------
        decision_vector_1d : array_like
            A 1d decision vector.

        Returns
        -------
        sequence
            The fitness of the input decision vector (concatenating the objectives,
            the equality and the inequality constraints)
        """
        # TODO: Fix this
        if self.pop is None:
            raise NotImplementedError("'pop' is not initialized.")

        try:

            # TODO: Use directory 'logging.'
            logger = logging.getLogger("pyxel")
            prev_log_level = logger.getEffectiveLevel()

            parameter_1d = self.convert_to_parameters(decision_vector_1d)
            # TODO: deepcopy is not needed. Check this
            processor_list = self.param_processor_list  # type: t.Sequence[Processor]

            overall_fitness = 0.0  # type: float
            for i, (processor, target_data) in enumerate(
                zip(processor_list, self.all_target_data)
            ):

                processor = self.update_processor(
                    parameter=parameter_1d, processor=processor
                )

                logger.setLevel(logging.WARNING)
                # result_proc = None
                if self.calibration_mode == CalibrationMode.Pipeline:
                    _ = run_exposure_pipeline(processor=processor, readout=self.readout)
                # elif self.calibration_mode == 'single_model':
                #     self.fitted_model.function(processor.detector)               # todo: update
                else:
                    raise NotImplementedError

                logger.setLevel(prev_log_level)

                simulated_data = self.get_simulated_data(processor=processor)

                weighting = None  # type: t.Optional[np.ndarray]

                if self.weighting is not None:
                    weighting = self.weighting[i] * np.ones(
                        (
                            processor.detector.geometry.row,
                            processor.detector.geometry.col,
                        )
                    )
                elif self.weighting_from_file is not None:
                    weighting = self.weighting_from_file[i]

                overall_fitness += self.calculate_fitness(
                    simulated_data=simulated_data,
                    target_data=target_data,
                    weighting=weighting,
                )

        except Exception:
            logging.exception(
                "Catch an exception in 'fitness' for ModelFitting: %r. exc: %r",
                self,
            )
            raise

        return [overall_fitness]

    def convert_to_parameters(self, decisions_vector: "ArrayLike") -> np.ndarray:
        """Convert a decision version from Pygmo2 to parameters.

        Parameters
        ----------
        decisions_vector : array_like
            It could a 1D or 2D array.

        Returns
        -------
        array_like
            Parameters
        """
        parameters = np.asarray(decisions_vector)

        a = 0
        for var in self.variables:
            b = 1
            if isinstance(var.values, list):
                b = len(var.values)
            if var.logarithmic:
                start = a
                stop = a + b
                parameters[..., start:stop] = np.power(10, parameters[..., start:stop])
            a += b

        return parameters

    # # TODO: Check this
    # def apply_parameters(
    #     self, processor: Processor, parameter: np.ndarray
    # ) -> Processor:
    #     """Create a new ``Processor`` with new parameters."""
    #     new_processor = self.update_processor(parameter=parameter, processor=processor)
    #
    #     if self.calibration_mode is CalibrationMode.Pipeline:
    #         new_processor.run_pipeline()
    #
    #     return new_processor

    def apply_parameters(
        self, processor: Processor, parameter: np.ndarray
    ) -> Processor:
        """Create a new ``Processor`` with new parameters."""
        new_processor = self.update_processor(parameter=parameter, processor=processor)

        _ = run_exposure_pipeline(processor=new_processor, readout=self.readout)

        return new_processor

    # # TODO: Check this
    # def apply_parameters_to_processors(self, parameters: xr.DataArray) -> pd.DataFrame:
    #     """TBW."""
    #     assert "island" in parameters.dims
    #     assert "param_id" in parameters.dims
    #
    #     lst = []
    #     for id_processor, processor in enumerate(self.param_processor_list):
    #         delayed_processor = delayed(processor)
    #
    #         for idx_island, params_array in parameters.groupby("island"):
    #             params = params_array.data  # type: np.ndarray
    #
    #             new_processor = delayed(self.apply_parameters)(
    #                 processor=delayed_processor, parameter=params
    #             )
    #
    #             lst.append(
    #                 {
    #                     "island": idx_island,
    #                     "id_processor": id_processor,
    #                     "processor": new_processor,
    #                 }
    #             )
    #
    #     df = pd.DataFrame(lst).sort_values(["island", "id_processor"])
    #
    #     return df

    def apply_parameters_to_processors(
        self, parameters: "xr.DataArray"
    ) -> pd.DataFrame:
        """TBW."""
        assert "island" in parameters.dims
        assert "param_id" in parameters.dims

        lst = []
        for id_processor, processor in enumerate(self.param_processor_list):
            delayed_processor = delayed(processor)

            for idx_island, params_array in parameters.groupby("island"):
                params = params_array.data  # type: np.ndarray

                result_processor = delayed(self.apply_parameters)(
                    processor=delayed_processor, parameter=params
                )

                lst.append(
                    {
                        "island": idx_island,
                        "id_processor": id_processor,
                        "processor": result_processor,
                    }
                )

        df = pd.DataFrame(lst).sort_values(["island", "id_processor"])

        return df

    def update_processor(
        self, parameter: np.ndarray, processor: Processor
    ) -> Processor:
        """TBW."""
        new_processor = copy.deepcopy(processor)
        a, b = 0, 0
        for var in self.variables:
            if var.values == "_":
                b = 1
                new_processor.set(key=var.key, value=parameter[a])
            elif isinstance(var.values, list):
                b = len(var.values)

                start = a
                stop = a + b
                new_processor.set(key=var.key, value=parameter[start:stop])
            a += b
        return new_processor

    # def get_results(
    #     self, overall_fitness: np.ndarray, parameter: np.ndarray
    # ) -> CalibrationResult:
    #     """TBW.
    #
    #     :param overall_fitness:
    #     :param parameter:
    #     :return:
    #     """
    #     results = {}  # type: t.Dict[str, t.Union[int, float]]
    #     fitness = overall_fitness[0]  # type: float
    #
    #     # TODO: Apply a copy of 'parameter' in 'self.old_update_parameter' ??
    #     parameter = self.old_update_parameter(parameter)
    #
    #     if self.file_path:
    #         ref_values_1d = np.concatenate([overall_fitness, parameter])
    #
    #         for island_num, value in self.match.items():
    #             values_1d = np.asarray(value, dtype=np.float)
    #
    #             if np.allclose(ref_values_1d, values_1d):
    #                 island = island_num
    #                 break
    #         else:
    #             raise RuntimeError
    #     #     # TODO: Use 'np.ravel(parameter)' instead of 'parameter.reshape(1, len(parameter))' ?
    #     #     arr_2d = np.c_[
    #     #         overall_fitness, parameter.reshape(1, len(parameter))
    #     #     ]  # type: np.ndarray
    #     #
    #     #     with np.printoptions(formatter={"float": "{: .6E}".format}, suppress=False):
    #     #         arr_str = np.array2string(
    #     #             arr_2d, separator="", suppress_small=False
    #     #         )  # type: str
    #     #
    #     #     arr_str = (
    #     #         arr_str.replace("\n", "")
    #     #         .replace("   ", " ")
    #     #         .replace("  ", " ")
    #     #         .replace("[[ ", "")
    #     #         .replace("]]", "")
    #     #     )
    #     #
    #     #     lst_str = arr_str.split(" ")  # type: t.Sequence[str]
    #     #     island = -1  # type: int
    #     #     # TODO: Use np.allclose to compare
    #     #     for k, v in self.match.items():
    #     #         if lst_str == v:
    #     #             island = k
    #     #             break
    #     #     if island == -1:
    #     #         raise RuntimeError()
    #     else:
    #         island = 0
    #
    #     logging.info(
    #         "Post-processing island %d, champion fitness: %1.5e",
    #         island,
    #         fitness,
    #     )
    #
    #     champion_list = deepcopy(
    #         self.param_processor_list
    #     )  # type: t.Sequence[Processor]
    #     for processor in champion_list:
    #         processor = self.old_update_processor(
    #             parameter=parameter, new_processor=processor
    #         )
    #         if self.calibration_mode is CalibrationMode.Pipeline:
    #             processor.run_pipeline()
    #
    #     a, b = 0, 0
    #     for var in self.variables:
    #         if var.values == "_":
    #             b = 1
    #             results[var.key] = parameter[a]
    #         elif isinstance(var.values, list):
    #             b = len(var.values)
    #
    #             start = a
    #             stop = a + b
    #             results[var.key] = parameter[start:stop]
    #         a += b
    #
    #     return CalibrationResult(
    #         processors=champion_list, fitness=fitness, island=island, results=results
    #     )

    # def champion_to_file(self, parameter: np.ndarray) -> None:
    #     """Get champion of each generation and write it to output files together with last population.
    #
    #     :return:
    #     """
    #     if self.champion_f_list is None:
    #         raise RuntimeError(
    #             "'champion_f_list' was not initialized with method '.configure(...)'."
    #         )
    #
    #     if self.champion_x_list is None:
    #         raise RuntimeError(
    #             "'champion_x_list' was not initialized with method '.configure(...)'."
    #         )
    #
    #     if self.fitness_array is None:
    #         raise RuntimeError(
    #             "'fitness_array' was not initialized with method '.configure(...)'."
    #         )
    #
    #     if self.population is None:
    #         raise RuntimeError(
    #             "'population' was not initialized with method '.configure(...)'."
    #         )
    #
    #     best_index = np.argmin(self.fitness_array)  # type: int
    #
    #     if self.g == 0:
    #         self.champion_f_list[self.g] = self.fitness_array[best_index]
    #         self.champion_x_list[self.g] = self.population[best_index, :]
    #     else:
    #         best_champ_index = np.argmin(self.champion_f_list)
    #
    #         if self.fitness_array[best_index] < self.champion_f_list[best_champ_index]:
    #             self.champion_f_list = np.vstack(
    #                 (self.champion_f_list, self.fitness_array[best_index])
    #             )
    #             self.champion_x_list = np.vstack(
    #                 (self.champion_x_list, self.population[best_index])
    #             )
    #         else:
    #             self.champion_f_list = np.vstack(
    #                 (self.champion_f_list, self.champion_f_list[-1])
    #             )
    #             self.champion_x_list = np.vstack(
    #                 (self.champion_x_list, self.champion_x_list[-1])
    #             )
    #
    #     # TODO: should we keep and write to file the population(s) which had the champion inside?
    #     # because usually this is not the last population currently we save to file!
    #
    #     if self.file_path and self.g > 0:
    #         self.add_to_champ_file(parameter)
    #         self.add_to_pop_file(parameter)

    # def save_population(self, parameter: np.ndarray, overall_fitness: float) -> None:
    #     """Save population of each generation to get champions.
    #
    #     :param parameter: 1d np.array
    #     :param overall_fitness: list
    #     :return:
    #     """
    #     if self.pop is None:
    #         raise RuntimeError("'pop' was not initialized with method '.configure'.")
    #
    #     if self.n % self.pop == 0:
    #         self.fitness_array = np.array([overall_fitness])
    #         self.population = parameter
    #     else:
    #         self.fitness_array = np.vstack(
    #             (self.fitness_array, np.array([overall_fitness]))
    #         )
    #         self.population = np.vstack((self.population, parameter))

    # def add_to_champ_file(self, parameter: np.ndarray) -> None:
    #     """TBW."""
    #     assert self.champion_f_list is not None
    #     assert self.champion_x_list is not None
    #     assert self.file_path
    #
    #     # TODO: Use `astropy.table` ?
    #     champions_file = self.file_path.joinpath(
    #         f"champions_id_{id(self)}.out"
    #     )  # type: Path
    #     str_format = "%d" + (len(parameter) + 1) * " %.6E"
    #     with champions_file.open("ab") as file1:
    #         np.savetxt(
    #             file1,
    #             np.c_[
    #                 np.array([self.g]),
    #                 self.champion_f_list[self.g],
    #                 self.champion_x_list[self.g, :].reshape(1, len(parameter)),
    #             ],
    #             fmt=str_format,
    #         )

    # def add_to_pop_file(self, parameter: np.ndarray) -> None:
    #     """TBW."""
    #     assert self.fitness_array is not None
    #     assert self.file_path
    #
    #     pop_file = self.file_path.joinpath(f"population_id_{id(self)}.out")
    #     str_format = "%d" + (len(parameter) + 1) * " %.6E"
    #     with pop_file.open("wb") as file2:
    #         np.savetxt(
    #             file2,
    #             np.c_[
    #                 self.g * np.ones(self.fitness_array.shape),
    #                 self.fitness_array,
    #                 self.population,
    #             ],
    #             fmt=str_format,
    #         )

    # def least_squares(self, simulated_data, dataset=None):
    #     """TBW.
    #
    #     :param simulated_data:
    #     :param dataset: int
    #     :return:
    #     """
    #     input_array = simulated_data[self.sim_fit_range]
    #
    #     if dataset is not None:
    #         if self.normalization:
    #             input_array = self.normalize(input_array, dataset=dataset)
    #             target = self.target_data_norm[dataset][self.targ_fit_range]
    #         else:
    #             target = self.target_data[dataset][self.targ_fit_range]
    #     else:
    #         if self.normalization:
    #             input_array = self.normalize(input_array)
    #             target = self.target_data_norm[self.targ_fit_range]
    #         else:
    #             target = self.target_data[self.targ_fit_range]
    #
    #     diff = target - input_array
    #     diff_square = diff * diff
    #     return np.sum(diff_square)

    # def set_normalization(self):
    #     """TBW.
    #
    #     :return:
    #     """
    #     self.normalization = True
    #     for i in range(len(self.target_data)):
    #         self.target_data_norm += [self.normalize(self.target_data[i], dataset=i)]

    # def normalize(self, array, dataset):
    #     """Normalize dataset arrays by injected signal maximum.
    #
    #     :param array: 1d np.array
    #     :param dataset: int
    #     :return:
    #     """
    #     return array / np.average(self.target_data[dataset][self.targ_fit_range])


# def merge_fitness(f):
#     """TBW."""
#     return f



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

import logging

from pyxel.detectors import MKID


def dead_time_filter(detector: MKID, dead_time: float) -> None:
    """TBW."""
    logging.info("")

    maximum_count = 1.0 / dead_time

    detector.phase.array[detector.phase.array >= maximum_count] = maximum_count



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Simple model to load charge profiles."""

import logging
import typing as t
from functools import lru_cache
from pathlib import Path

import numpy as np
import pandas as pd

from pyxel.data_structure import Charge
from pyxel.detectors import Detector, Geometry
from pyxel.detectors.geometry import (
    get_horizontal_pixel_center_pos,
    get_vertical_pixel_center_pos,
)


@lru_cache(maxsize=128)  # One must add parameter 'maxsize' for Python 3.7
def _create_charges(
    num_rows: int,
    num_cols: int,
    pixel_vertical_size: float,
    pixel_horizontal_size: float,
    txt_file: str,
    profile_position_y: int,
    profile_position_x: int,
    fit_profile_to_det: bool = False,
) -> pd.DataFrame:
    """Create charges from a charge profile file."""
    # All pixels has zero charge by default
    detector_charge_2d = np.zeros((num_rows, num_cols))

    # Load 2d charge profile (which can be smaller or
    #                         larger in dimensions than detector imaging area)
    full_path = Path(txt_file).resolve()
    charges_from_file_2d = np.loadtxt(str(full_path), ndmin=2)  # type: np.ndarray

    if fit_profile_to_det:
        # Crop 2d charge profile, so it is not larger in dimensions than detector imaging area)
        charges_from_file_2d = charges_from_file_2d[
            slice(0, num_rows), slice(0, num_cols)
        ]

    profile_rows, profile_cols = charges_from_file_2d.shape

    detector_charge_2d[
        slice(profile_position_y, profile_position_y + profile_rows),
        slice(profile_position_x, profile_position_x + profile_cols),
    ] = charges_from_file_2d

    charge_numbers = detector_charge_2d.flatten()  # type: np.ndarray
    where_non_zero = np.where(charge_numbers > 0.0)
    charge_numbers = charge_numbers[where_non_zero]
    size = charge_numbers.size  # type: int

    vertical_pixel_center_pos_1d = get_vertical_pixel_center_pos(
        num_rows=num_rows,
        num_cols=num_cols,
        pixel_vertical_size=pixel_vertical_size,
    )

    horizontal_pixel_center_pos_1d = get_horizontal_pixel_center_pos(
        num_rows=num_rows,
        num_cols=num_cols,
        pixel_horizontal_size=pixel_horizontal_size,
    )

    init_ver_pix_position_1d = vertical_pixel_center_pos_1d[where_non_zero]
    init_hor_pix_position_1d = horizontal_pixel_center_pos_1d[where_non_zero]

    # Create new charges
    return Charge.create_charges(
        particle_type="e",
        particles_per_cluster=charge_numbers,
        init_energy=np.zeros(size),
        init_ver_position=init_ver_pix_position_1d,
        init_hor_position=init_hor_pix_position_1d,
        init_z_position=np.zeros(size),
        init_ver_velocity=np.zeros(size),
        init_hor_velocity=np.zeros(size),
        init_z_velocity=np.zeros(size),
    )


# TODO: Fix this
# @validators.validate
# @config.argument(name='txt_file', label='file path', units='', validate=checkers.check_path)
def charge_profile(
    detector: Detector,
    txt_file: t.Union[str, Path],
    fit_profile_to_det: bool = False,
    profile_position: t.Optional[t.Tuple[int, int]] = None,
) -> None:
    """Load charge profile from txt file for detector, mostly for but not limited to CCDs.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    txt_file : str or Path
        File path.
    fit_profile_to_det : bool
    profile_position : list
    """
    logging.info("")

    if profile_position is None:
        profile_position_y = 0  # type: int
        profile_position_x = 0  # type: int
    else:
        profile_position_y, profile_position_x = profile_position

    geo = detector.geometry  # type: Geometry

    # Create charges as `DataFrame`
    charges = _create_charges(
        num_rows=geo.row,
        num_cols=geo.col,
        pixel_vertical_size=geo.pixel_vert_size,
        pixel_horizontal_size=geo.pixel_horz_size,
        txt_file=txt_file,
        profile_position_y=profile_position_y,
        profile_position_x=profile_position_x,
        fit_profile_to_det=fit_profile_to_det,
    )  # type: pd.DataFrame

    # Add charges in 'detector'
    detector.charge.add_charge_dataframe(charges)



#  Copyright (c) YOUR COPYRIGHT HERE

"""
persistence module for the PyXel simulation.

This module is used in charge_collection

-------------------------------------------------------------------------------

+--------------+----------------------------------+---------------------------+
| Author       | Name                             | Creation                  |
+--------------+----------------------------------+---------------------------+
| You          | charge_collection                | 05/20/2021                |
+--------------+----------------------------------+---------------------------+

+-----------------+-------------------------------------+---------------------+
| Contributor     | Name                                | Creation            |
+-----------------+-------------------------------------+---------------------+
| Name            | filename                            | -                   |
+-----------------+-------------------------------------+---------------------+

This module can be found in pyxel/models/charge_collection/persistence.py

Algorithm
=========

The persistence model take as input the total number of detrapped charges
(persistence map with a 10000 seconds soak at more than the FWC) and a map
which is the previous one divided by full well capacity. This gives the total
amount of trap per charge.

At each iteration of the pipeline, the model will first check if there is a
'persistence' entry in the memore of the detector and if not will create it
(happens only at the first iteration).
Then, it will compute the amount of trapped charges in this iteration, add it
to the memory of the detector and then will remove this amount from the pixel array.

Default parameter
=================

The defaults parameter are the one derived from the characterization of the ENG grade
detector for MOONS (H4RG).

+----------------+-------------+
| Time constants | Proportions |
+================+=============+
| 1              | 0.307       |
+----------------+-------------+
| 10             | 0.175       |
+----------------+-------------+
| 100            | 0.188       |
+----------------+-------------+
| 1000           | 0.136       |
+----------------+-------------+
| 10000          | 0.194       |
+----------------+-------------+

Code example
============

.. code-block:: python

    # To access the memory of the detector with the persistence map

    detector._memory["persistence"]

    # This dictionnary consist of 5 entries (one per time constant)

.. literalinclude:: pyxel/models/charge_collection/persistence.py
    :language: python
    :linenos:
    :lines: 84-87

Model reference in the YAML config file
=======================================

.. code-block:: yaml

    pipeline:

      # Persistence model based on MOONS detector (H4RG) measurements
      - name: persistence
        func: pyxel.models.charge_collection.persistence.current_persistence
        enabled: true
        arguments:
          trap_timeconstants: [1, 10, 100, 1000, 10000]
          trap_densities: data/fits/20210408121614_20210128_ENG20370_AUTOCHAR-Persistence_FitTrapDensityMap.fits
          trap_max: data/fits/20210408093114_20210128_ENG20370_AUTOCHAR-Persistence_FitMaximumTrapMap.fits
          trap_proportions: [0.307, 0.175, 0.188, 0.136, 0.194]

Useful links
============

Persistence paper from S. Tulloch
https://arxiv.org/abs/1908.06469

ReadTheDocs documentation
https://sphinx-rtd-theme.readthedocs.io/en/latest/index.html

.. todo::

   - Add temperature dependency
   - Add default -flat?- trap maps

"""

import logging

import numpy as np
from astropy.io import fits

from pyxel.detectors import CMOS


# @validators.validate
# @config.argument(name='', label='', units='', validate=)
def simple_persistence(
    detector: CMOS, trap_timeconstants: list, trap_densities: list
) -> None:
    """Trapping/detrapping charges."""
    logging.info("Persistence")
    if "persistence" not in detector._memory.keys():
        detector._memory["persistence"] = dict()
        for trap_density, trap_timeconstant in zip(trap_densities, trap_timeconstants):
            entry = "".join(
                ["trappedCharges_", str(trap_density), "-", str(trap_timeconstant)]
            )
            detector._memory["persistence"].update(
                {entry: np.zeros((detector.geometry.row, detector.geometry.col))}
            )
    else:
        for trap_density, trap_timeconstant in zip(trap_densities, trap_timeconstants):
            entry = "".join(
                ["trappedCharges_", str(trap_density), "-", str(trap_timeconstant)]
            )
            trapped_charges = detector._memory["persistence"][entry]
            # Trap density is a scalar for now, in the future we could feed maps?
            # the delta t is fixed to 0.5 s, need to find a way to avoid problem of divergence
            trapped_charges = trapped_charges + (0.5 / trap_timeconstant) * (
                detector.pixel.array * trap_density - trapped_charges
            )
            # Remove the trapped charges from the pixel
            detector.pixel.array -= trapped_charges.astype(np.int32)
            # Replace old trapped charges map in the detector's memory
            detector._memory["persistence"][entry] = trapped_charges

    return None


def current_persistence(
    detector: CMOS,
    trap_timeconstants: list,
    trap_densities: str,
    trap_max: str,
    trap_proportions: list,
) -> None:
    """Trapping/detrapping charges."""
    logging.info("Persistence")
    # If the file for trap density is correct open it and use it
    # otherwise I need to define a default trap density map

    # Extract trap density / full well
    trap_densities_2d = fits.open(trap_densities)[0].data[
        : detector.geometry.row, : detector.geometry.col
    ]  # type: np.ndarray
    trap_densities_2d[np.where(trap_densities_2d < 0)] = 0

    # Extract the max amount of trap by long soak
    trap_max_2d = fits.open(trap_max)[0].data[
        : detector.geometry.row, : detector.geometry.col
    ]
    trap_max_2d[np.where(trap_max_2d < 0)] = 0

    # If there is no entry for persistence in the memory of the detector
    # create one
    if "persistence" not in detector._memory.keys():
        detector._memory["persistence"] = dict()
        for trap_proportion, trap_timeconstant in zip(
            trap_proportions, trap_timeconstants
        ):
            entry = "".join(
                ["trappedCharges_", str(trap_proportion), "-", str(trap_timeconstant)]
            )
            detector._memory["persistence"].update(
                {entry: np.zeros((detector.geometry.row, detector.geometry.col))}
            )
            trapped_charges = detector._memory["persistence"][entry]

    # For each trap population
    for trap_proportion, trap_timeconstant in zip(trap_proportions, trap_timeconstants):

        # Get the correct persistence traps entry
        entry = "".join(
            ["trappedCharges_", str(trap_proportion), "-", str(trap_timeconstant)]
        )

        # Select the trapped charges array
        trapped_charges = detector._memory["persistence"][entry]

        # Time for reading a frame
        # delta_t = (detector.geometry.row * detector.geometry.col)/detector.characteristics.readout_freq
        delta_t = detector.time_step

        # Computer trapped charge for this increament of time
        # Time factor is the integration time divided by the time constant (1, 10, 100, 1000, 10000)
        time_factor = delta_t / trap_timeconstant

        # Amount of charges trapped per unit of full well
        max_charges = trap_densities_2d * trap_proportion

        # Maximum of amount of charges trapped
        fw_trap = trap_max_2d * trap_proportion

        diff = time_factor * (
            max_charges * detector.pixel.array * np.exp(-time_factor) - trapped_charges
        )
        # Compute trapped charges
        trapped_charges = trapped_charges + time_factor * (
            max_charges * detector.pixel.array * np.exp(-time_factor) - trapped_charges
        )

        # When the amount of trapped charges is superior to the maximum of available traps, set to max
        trapped_charges[np.where(trapped_charges > fw_trap)] = max_charges[
            np.where(trapped_charges > fw_trap)
        ]
        # Can't have a negative amount of charges trapped
        trapped_charges[np.where(trapped_charges < 0)] = 0

        # Remove the trapped charges from the pixel
        # detector.pixel.array -= trapped_charges.astype(np.int32)
        detector.pixel.array -= diff.astype(np.int32)

        # Replace old trapped charges map in the detector's memory
        detector._memory["persistence"][entry] = trapped_charges

    return None



from collections import abc
from pathlib import Path

import pytest

import pyxel
from pyxel import Configuration
from pyxel.detectors import CCD
from pyxel.observation.observation import Observation, ParameterMode
from pyxel.pipelines import DetectionPipeline, Processor

try:
    import pygmo as pg

    WITH_PYGMO = True
except ImportError:
    WITH_PYGMO = False


expected_sequential = [
    (0, [("level", 10), ("initial_energy", 100)]),
    (1, [("level", 20), ("initial_energy", 100)]),
    (2, [("level", 30), ("initial_energy", 100)]),
    (3, [("level", 100), ("initial_energy", 100)]),
    (4, [("level", 100), ("initial_energy", 200)]),
    (5, [("level", 100), ("initial_energy", 300)]),
]

expected_product = [
    (0, [("level", 10), ("initial_energy", 100)]),
    (1, [("level", 10), ("initial_energy", 200)]),
    (2, [("level", 10), ("initial_energy", 300)]),
    (3, [("level", 20), ("initial_energy", 100)]),
    (4, [("level", 20), ("initial_energy", 200)]),
    (5, [("level", 20), ("initial_energy", 300)]),
    (6, [("level", 30), ("initial_energy", 100)]),
    (7, [("level", 30), ("initial_energy", 200)]),
    (8, [("level", 30), ("initial_energy", 300)]),
]


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "mode, expected",
    [
        # ('single', expected_single),
        (ParameterMode.Sequential, expected_sequential),
        (ParameterMode.Product, expected_product),
    ],
)
def test_pipeline_parametric_without_init_photon(mode: ParameterMode, expected):
    input_filename = "tests/data/parametric.yaml"
    cfg = pyxel.load(Path(input_filename))

    assert isinstance(cfg, Configuration)
    assert hasattr(cfg, "observation")
    assert hasattr(cfg, "ccd_detector")
    assert hasattr(cfg, "pipeline")

    observation = cfg.observation
    assert isinstance(observation, Observation)

    observation.parameter_mode = mode

    detector = cfg.ccd_detector
    assert isinstance(detector, CCD)

    assert detector.has_photon is False
    with pytest.raises(
        RuntimeError,
        match=(
            r"Photon array is not initialized ! "
            r"Please use a 'Photon Generation' model"
        ),
    ):
        _ = detector.photon

    pipeline = cfg.pipeline
    assert isinstance(pipeline, DetectionPipeline)

    processor = Processor(detector=detector, pipeline=pipeline)
    result = observation.debug_parameters(processor)
    assert result == expected

    processor_generator = observation._processors_it(processor=processor)
    assert isinstance(processor_generator, abc.Generator)

    for proc, _, _ in processor_generator:
        assert isinstance(proc, Processor)

        proc.run_pipeline()



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

import typing as t
from enum import Enum
from pathlib import Path

import numpy as np
import pandas as pd

from pyxel import load_datacube, load_image

if t.TYPE_CHECKING:
    import xarray as xr

#  from pyxel.pipelines import Processor
__all__ = [
    "CalibrationResult",
    "CalibrationMode",
    "Island",
    "check_ranges",
    "list_to_slice",
    "read_data",
]


# @attr.s(frozen=True, auto_attribs=True, slots=True)
# class CalibrationResult:
#     """TBW."""
#
#     processors: t.Sequence[Processor]
#     fitness: float
#     island: int
#     results: t.Mapping[str, t.Union[int, float]]


class CalibrationResult(t.NamedTuple):
    """Result class for calibration class."""

    dataset: "xr.Dataset"
    processors: pd.DataFrame
    logs: pd.DataFrame
    filenames: t.Sequence


class CalibrationMode(Enum):
    """TBW."""

    Pipeline = "pipeline"
    SingleModel = "single_model"


class Island(Enum):
    """User defines num_islands provides by Pygmo."""

    MultiProcessing = "multiprocessing"
    MultiThreading = "multithreading"
    IPyParallel = "ipyparallel"


def read_single_data(filename: Path) -> np.ndarray:
    """Read a numpy array from a FITS or NPY file.

    Parameters
    ----------
    filename : Path

    Returns
    -------
    array
        TBW.
    """
    # Late import to avoid circular import

    data = load_image(filename)

    # # TODO: Is it the right manner ?
    # if data is None:
    #     raise OSError(f"Input file '{filename}' can not be read by Pyxel.")

    return data


def read_single_datacube(filename: Path) -> np.ndarray:
    """Read a numpy array from a FITS or NPY file.

    Parameters
    ----------
    filename : Path

    Returns
    -------
    array
        TBW.
    """

    data = load_datacube(filename)

    # # TODO: Is it the right manner ?
    # if data is None:
    #     raise OSError(f"Input file '{filename}' can not be read by Pyxel.")

    return data


def read_datacubes(filenames: t.Sequence[Path]) -> t.Sequence[np.ndarray]:
    """Read numpy array(s) from several FITS or NPY files.

    :param filenames:
    :return:
    """
    output = [
        read_single_datacube(Path(filename)) for filename in filenames
    ]  # type: t.Sequence[np.ndarray]

    return output


def read_data(filenames: t.Sequence[Path]) -> t.Sequence[np.ndarray]:
    """Read numpy array(s) from several FITS or NPY files.

    :param filenames:
    :return:
    """
    output = [
        read_single_data(Path(filename)) for filename in filenames
    ]  # type: t.Sequence[np.ndarray]

    return output


# TODO: Create unit tests for this function
def list_to_slice(
    input_list: t.Optional[t.Sequence[int]] = None,
) -> t.Union[t.Tuple[slice, slice], t.Tuple[slice, slice, slice]]:
    """TBW.

    :return:
    """
    if not input_list:
        return slice(None), slice(None)

    elif len(input_list) == 4:
        return slice(input_list[0], input_list[1]), slice(input_list[2], input_list[3])

    elif len(input_list) == 6:
        return (
            slice(input_list[0], input_list[1]),
            slice(input_list[2], input_list[3]),
            slice(input_list[4], input_list[5]),
        )

    else:
        raise ValueError("Fitting range should have 4 or 6 values")


# TODO: Create unit tests for this function
def list_to_3d_slice(
    input_list: t.Optional[t.Sequence[int]] = None,
) -> t.Tuple[slice, slice, slice]:
    """TBW.

    :return:
    """
    if not input_list:
        return slice(None), slice(None), slice(None)

    elif len(input_list) == 6:
        return (
            slice(input_list[0], input_list[1]),
            slice(input_list[2], input_list[3]),
            slice(input_list[4], input_list[5]),
        )

    else:
        raise ValueError("Fitting range should have 6 values")


def check_ranges(
    target_fit_range: t.Sequence[int],
    out_fit_range: t.Sequence[int],
    rows: int,
    cols: int,
    readout_times: t.Optional[int] = None,
) -> None:
    """TBW."""
    if target_fit_range:
        if len(target_fit_range) not in (4, 6):
            raise ValueError

        if out_fit_range:
            if len(out_fit_range) not in (4, 6):
                raise ValueError

            if (target_fit_range[1] - target_fit_range[0]) != (
                out_fit_range[1] - out_fit_range[0]
            ):
                raise ValueError(
                    "Fitting ranges have different lengths in 1st dimension"
                )

            # TODO: It could be refactor in a more pythonic way
            if len(target_fit_range) == 4 and len(out_fit_range) == 4:
                if (target_fit_range[3] - target_fit_range[2]) != (
                    out_fit_range[3] - out_fit_range[2]
                ):
                    raise ValueError(
                        "Fitting ranges have different lengths in 2nd dimension"
                    )

            if len(target_fit_range) == 6 and len(out_fit_range) == 6:
                if (target_fit_range[5] - target_fit_range[4]) != (
                    out_fit_range[5] - out_fit_range[4]
                ):
                    raise ValueError(
                        "Fitting ranges have different lengths in third dimension"
                    )

        # for i in [0, 1]:
        #     # TODO: It could be refactor in a more pythonic way
        #     if not (0 <= target_fit_range[i] <= rows):
        #         raise ValueError("Value of target fit range is wrong")

        for i in [-3, -4]:
            if not (0 <= target_fit_range[i] <= rows):
                raise ValueError("Value of target fit range is wrong")

        for i in [-1, -2]:
            if not (0 <= target_fit_range[i] <= cols):
                raise ValueError("Value of target fit range is wrong")

        if len(target_fit_range) == 6:
            if readout_times is None:
                raise ValueError("Target data is not a 3 dimensional array")

            for i in [0, 1]:
                # TODO: It could be refactor in a more pythonic way (this is optional)
                if not (0 <= target_fit_range[i] <= readout_times):
                    raise ValueError("Value of target fit range is wrong")
